#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import time

import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import CUSTOM_MODEL_IMAGE_TYPE, DEFAULT_MAX_WAIT, NETWORK_EGRESS_POLICY
from datarobot.errors import AsyncProcessUnsuccessfulError
from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


class CustomModelTest(APIObject):
    """An custom model test.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        test id
    custom_model_image_id: str
        id of a custom model image
    image_type: str
        the type of the image, either CUSTOM_MODEL_IMAGE_TYPE.CUSTOM_MODEL_IMAGE if the testing
        attempt is using a CustomModelImage as its model or
        CUSTOM_MODEL_IMAGE_TYPE.CUSTOM_MODEL_VERSION if the testing attempt is
        using a CustomModelVersion with dependency management
    overall_status: str
        a string representing testing status.
        Status can be
        - 'not_tested': the check not run
        - 'failed': the check failed
        - 'succeeded': the check succeeded
        - 'warning': the check resulted in a warning, or in non-critical failure
        - 'in_progress': the check is in progress
    detailed_status: dict
        detailed testing status - maps the testing types to their status and message.
        The keys of the dict are one of 'errorCheck', 'nullValueImputation',
        'longRunningService', 'sideEffects'.
        The values are dict with 'message' and 'status' keys.
    created_by: str
        a user who created a test
    dataset_id: Optional[str]
        id of a dataset used for testing
    dataset_version_id: Optional[str]
        id of a dataset version used for testing
    completed_at: Optional[str]
        ISO-8601 formatted timestamp of when the test has completed
    created_at: Optional[str]
        ISO-8601 formatted timestamp of when the version was created
    network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
        Determines whether the given custom model is isolated, or can access the public network.
        Values: [`datarobot.NETWORK_EGRESS_POLICY.NONE`, `datarobot.NETWORK_EGRESS_POLICY.PUBLIC`].
    maximum_memory: Optional[int]
        The maximum memory that might be allocated by the custom-model.
        If exceeded, the custom-model will be killed by k8s
    replicas: Optional[int]
        A fixed number of replicas that will be deployed in the cluster
    """

    _path = "customModelTests/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("custom_model_image_id"): String(),
            t.Key("image_type"): t.Enum(*CUSTOM_MODEL_IMAGE_TYPE.ALL),
            t.Key("overall_status"): String(),
            t.Key("testing_status")
            >> "detailed_status": t.Dict(
                {
                    t.Key(test_type): t.Dict(
                        {t.Key("status"): String(), t.Key("message"): String(allow_blank=True)}
                    ).allow_extra("*")
                    for test_type in [
                        "error_check",
                        "null_value_imputation",
                        "long_running_service",
                        "side_effects",
                    ]
                }
            ).allow_extra("*"),
            t.Key("created_by"): String(),
            t.Key("dataset_id", optional=True): String(),
            t.Key("dataset_version_id", optional=True): String(),
            t.Key("completed_at", optional=True): String(allow_blank=True),
            t.Key("created", optional=True) >> "created_at": String(),
            t.Key("network_egress_policy", optional=True): t.Enum(*NETWORK_EGRESS_POLICY.ALL),
            t.Key("maximum_memory", optional=True): Int(),
            t.Key("replicas", optional=True): Int(),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.id!r})"

    def _set_values(  # pylint: disable=missing-function-docstring
        self,
        id,
        custom_model_image_id,
        image_type,
        overall_status,
        detailed_status,
        created_by,
        dataset_id=None,
        dataset_version_id=None,
        completed_at=None,
        created_at=None,
        network_egress_policy=None,
        maximum_memory=None,
        replicas=None,
    ):
        self.id = id
        self.dataset_id = dataset_id
        self.dataset_version_id = dataset_version_id
        self.custom_model_image_id = custom_model_image_id
        self.image_type = image_type
        self.overall_status = overall_status
        self.detailed_status = detailed_status
        self.created_by = created_by
        self.completed_at = completed_at
        self.created_at = created_at
        self.network_egress_policy = network_egress_policy
        self.maximum_memory = maximum_memory
        self.replicas = replicas

    @classmethod
    def create(
        cls,
        custom_model_id,
        custom_model_version_id,
        dataset_id=None,
        max_wait=DEFAULT_MAX_WAIT,
        network_egress_policy=None,
        maximum_memory=None,
        replicas=None,
    ):
        """Create and start a custom model test.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        custom_model_version_id: str
            the id of the custom model version
        dataset_id: Optional[str]
            The id of the testing dataset for non-unstructured custom models.
            Ignored and not required for unstructured models.
        max_wait: Optional[int]
            max time to wait for a test completion.
            If set to None - method will return without waiting.
        network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
            Determines whether the given custom model is isolated, or can access the public network.
            Values: [`datarobot.NETWORK_EGRESS_POLICY.NONE`, `datarobot.NETWORK_EGRESS_POLICY.PUBLIC`].
        maximum_memory: Optional[int]
            The maximum memory that might be allocated by the custom-model.
            If exceeded, the custom-model will be killed by k8s
        replicas: Optional[int]
            A fixed number of replicas that will be deployed in the cluster

        Returns
        -------
        CustomModelTest
            created custom model test

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = {
            "custom_model_id": custom_model_id,
            "custom_model_version_id": custom_model_version_id,
        }

        if dataset_id:
            payload["dataset_id"] = dataset_id
        if network_egress_policy:
            payload["network_egress_policy"] = network_egress_policy
        if maximum_memory:
            payload["maximum_memory"] = maximum_memory
        if replicas:
            payload["replicas"] = replicas

        response = cls._client.post(cls._path, data=payload)

        # at this point custom model test is already created
        custom_model_test = cls.list(custom_model_id)[0]

        if max_wait is None:
            # return without waiting for the test to finish
            return custom_model_test
        else:
            try:
                # wait for the test to finish
                custom_model_test_loc = wait_for_async_resolution(
                    cls._client, response.headers["Location"], max_wait
                )
                return cls.from_location(custom_model_test_loc)
            except AsyncProcessUnsuccessfulError:
                # if the job was aborted server sends appropriate status and
                # `wait_for_async_resolution` raises exception,
                # but the test has been already created, and contains error log,
                # so return the test

                # the test needs some time to update its state
                max_state_wait = 10
                custom_model_test.refresh()

                start_time = time.time()
                while custom_model_test.overall_status == "in_progress":
                    if time.time() >= start_time + max_state_wait:
                        raise
                    time.sleep(1)
                    custom_model_test.refresh()

                return custom_model_test

    @classmethod
    def list(cls, custom_model_id):
        """List custom model tests.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model

        Returns
        -------
        List[CustomModelTest]
            a list of custom model tests

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = {"custom_model_id": custom_model_id}
        data = unpaginate(cls._path, payload, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, custom_model_test_id):
        """Get custom model test by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_test_id: str
            the id of the custom model test

        Returns
        -------
        CustomModelTest
            retrieved custom model test

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = f"{cls._path}{custom_model_test_id}/"
        return cls.from_location(path)

    def get_log(self):
        """Get log of a custom model test.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = f"{self._path}{self.id}/log/"
        return self._client.get(path).text

    def get_log_tail(self):
        """Get log tail of a custom model test.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = f"{self._path}{self.id}/tail/"
        return self._client.get(path).text

    def cancel(self):
        """Cancel custom model test that is in progress.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = f"{self._path}{self.id}/"
        self._client.delete(path)

    def refresh(self):
        """Update custom model test with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = f"{self._path}{self.id}/"

        response = self._client.get(path)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, cast, Dict, List, Optional, Union

import trafaret as t

from datarobot.enums import ENTITY_TYPES, INSIGHTS_SOURCES
from datarobot.insights.base import BaseInsight
from datarobot.utils import deprecation


class ShapImpact(BaseInsight):
    """Class for SHAP Impact calculations. Use the standard methods of BaseInsight to compute
    and retrieve: compute, create, list, get.
    """

    INSIGHT_NAME = "shapImpact"
    INSIGHT_DATA = {
        "shap_impacts": t.Or(
            t.List(t.List(t.Or(t.Int(), t.Float()))),
            t.Dict(
                {
                    t.Key("featureName"): t.String(),
                    t.Key("impactNormalized"): t.Float(),
                    t.Key("impactUnnormalized"): t.Float(),
                }
            ),
        ),
        "base_value": t.List(t.Float()),
        "capping": Optional[
            t.Or(
                t.Null(),
                t.Dict(
                    {
                        t.Key("right"): t.Or(t.String(), t.Float(), t.Null()),  # noqa: F821
                        t.Key("left"): t.Or(t.String(), t.Float(), t.Null()),  # noqa: F821
                    }
                ),
            )
        ],
        "link": t.Or(t.String(), t.Null()),
        "row_count": Optional[t.Int()],
    }

    @classmethod
    def _get_payload(
        cls,
        entity_id: str,
        source: str = INSIGHTS_SOURCES.TRAINING,
        data_slice_id: Optional[str] = None,
        external_dataset_id: Optional[str] = None,
        entity_type: Optional[ENTITY_TYPES] = None,
        quick_compute: Optional[bool] = None,
        row_count: Optional[int] = None,
        **kwargs: Any,
    ) -> Dict[str, Union[str, int, bool]]:
        """Construct a payload for a compute request"""
        payload = super()._get_payload(
            entity_id=entity_id,
            source=source,
            data_slice_id=data_slice_id,
            external_dataset_id=external_dataset_id,
            entity_type=entity_type,
            quick_compute=quick_compute,
            **kwargs,
        )

        if row_count:
            payload["rowCount"] = row_count

        return payload

    # deal with shap-impact specific data structure to pass impacts list to UI component
    def _get_datarobot_ui_data(self) -> Dict[str, Any]:
        return {"data": self.shap_impacts}

    def sort(self, key_name: str = "-impact_normalized") -> None:
        """
        Sorts insights data by key name.

        :param key_name: item key name to sort data.
            One of 'feature_name', 'impact_normalized' or 'impact_unnormalized'.
            Starting with '-' reverses sort order. Default '-impact_normalized'
        """
        reverse = False
        if not isinstance(key_name, str):
            raise TypeError
        if len(key_name) > 0 and key_name[0] == "-":
            reverse = True
            key_name = key_name[1:]
        self.data["shap_impacts"].sort(key=lambda x: x[key_name], reverse=reverse)

    @property
    def shap_impacts(self) -> List[List[Any]]:
        """SHAP impact values

        Returns
        -------
        shap impacts
            A list of the SHAP impact values
        """
        return cast(List[List[Any]], self.data["shap_impacts"])

    @property
    def base_value(self) -> List[float]:
        """A list of base prediction values"""
        return cast(List[float], self.data["base_value"])

    @property
    def capping(self) -> Optional[Dict[str, Any]]:
        """Capping for the models in the blender"""
        return cast(Optional[Dict[str, Any]], self.data.get("capping"))

    @property
    def link(self) -> Optional[str]:
        """Shared link function of the models in the blender"""
        return cast(Optional[str], self.data.get("link"))

    @property
    @deprecation.deprecated(
        deprecated_since_version="v3.6",
        will_remove_version="v3.7",
        message="row_count is deprecated.",
    )
    def row_count(self) -> Optional[int]:
        """Number of SHAP impact rows. This is deprecated."""
        return cast(Optional[int], self.data.get("row_count"))



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.genai.custom_model_embedding_validation import CustomModelEmbeddingValidation
from datarobot.models.genai.custom_model_validation import (
    get_entity_id,
    NonChatAwareCustomModelValidation,
)
from datarobot.models.genai.playground import Playground
from datarobot.models.use_cases.use_case import UseCase
from datarobot.models.use_cases.utils import get_use_case_id, resolve_use_cases, UseCaseLike
from datarobot.utils.pagination import unpaginate

chunking_parameters_trafaret = t.Dict(
    {
        t.Key("embedding_model", optional=True, default=None): str,
        t.Key("embedding_validation_id", optional=True, default=None): str,
        t.Key("chunking_method", optional=True, default=None): t.Or(str, t.Null),
        t.Key("chunk_size", optional=True, default=None): t.Or(t.Int, t.Null),
        t.Key("chunk_overlap_percentage", optional=True, default=None): t.Or(t.Int, t.Null),
        t.Key("separators", optional=True, default=None): t.Or(t.List(t.String), t.Null),
        t.Key("custom_chunking", default=False): t.Bool,
    }
).ignore_extra("*")

embedding_model_trafaret = t.Dict(
    {
        t.Key("embedding_model"): str,
        t.Key("description"): str,
        t.Key("max_sequence_length"): int,
        t.Key("languages"): str,
    }
).ignore_extra("*")

supported_custom_model_embedding_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String,
    }
).ignore_extra("*")

supported_embeddings_trafaret = t.Dict(
    {
        t.Key("embedding_models"): t.List(embedding_model_trafaret),
        t.Key("default_embedding_model"): t.String,
        t.Key("custom_model_embedding_validations"): t.List(
            supported_custom_model_embedding_trafaret
        ),
    }
).ignore_extra("*")

text_chunking_parameter_fields_trafaret = t.Dict(
    {
        t.Key("name"): t.String,
        t.Key("type"): t.String,
        t.Key("description"): t.String,
        t.Key("default"): t.Or(t.Int, t.List(t.String(allow_blank=True))),
        t.Key("min", optional=True): t.Or(t.Int, t.Null),
        t.Key("max", optional=True): t.Or(t.Int, t.Null),
    }
).ignore_extra("*")

text_chunking_method_trafaret = t.Dict(
    {
        t.Key("chunking_method"): t.Or(t.String, t.Null),
        t.Key("chunking_parameters"): t.List(text_chunking_parameter_fields_trafaret),
        t.Key("description"): t.String,
    }
).ignore_extra("*")

text_chunking_config_trafaret = t.Dict(
    {
        t.Key("embedding_model"): t.String,
        t.Key("methods"): t.List(text_chunking_method_trafaret),
        t.Key("default_method"): t.String,
    }
).ignore_extra("*")

supported_text_chunkings_trafaret = t.Dict(
    {
        t.Key("text_chunking_configs"): t.List(text_chunking_config_trafaret),
    }
).ignore_extra("*")

nested_supported_retrieval_setting_trafaret = t.Dict(
    {
        t.Key("name"): t.String,
        t.Key("type"): t.Or(t.String, t.List(t.String)),
        t.Key("description"): t.String,
        t.Key("title"): t.String,
        t.Key("default", optional=True, default=None): t.Or(t.String, t.Int, t.Bool, t.Null),
        t.Key("minimum", optional=True, default=None): t.Or(t.Int, t.Null),
        t.Key("maximum", optional=True, default=None): t.Or(t.Int, t.Null),
        t.Key("enum", optional=True, default=None): t.Or(t.List(t.String), t.Null),
        t.Key("settings", optional=True, default=None): t.Null,
        t.Key("group_id", optional=True, default=None): t.Or(t.String, t.Null),
    }
).ignore_extra("*")

supported_retrieval_setting_trafaret = t.Dict(
    {
        t.Key("name"): t.String,
        t.Key("type"): t.Or(t.String, t.List(t.String)),
        t.Key("description"): t.String,
        t.Key("title"): t.String,
        t.Key("default", optional=True, default=None): t.Or(t.String, t.Int, t.Bool, t.Null),
        t.Key("minimum", optional=True, default=None): t.Or(t.Int, t.Null),
        t.Key("maximum", optional=True, default=None): t.Or(t.Int, t.Null),
        t.Key("enum", optional=True, default=None): t.Or(t.List(t.String), t.Null),
        t.Key("settings", optional=True, default=None): t.Or(
            t.List(nested_supported_retrieval_setting_trafaret), t.Null
        ),
        t.Key("group_id", optional=True, default=None): t.Or(t.String, t.Null),
    }
).ignore_extra("*")

supported_retrieval_settings_trafaret = t.Dict(
    {
        t.Key("settings"): t.List(supported_retrieval_setting_trafaret),
    }
).ignore_extra("*")


vector_database_dataset_export_job_trafaret = t.Dict(
    {
        t.Key("job_id"): t.String,
        t.Key("vector_database_id"): t.String,
        t.Key("export_dataset_id"): t.String,
    }
).ignore_extra("*")

vector_database_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String,
        t.Key("size"): t.Int,
        t.Key("use_case_id"): t.String,
        t.Key("dataset_id", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("embedding_model", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("chunking_method", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("chunk_size", optional=True, default=None): t.Or(t.Int, t.Null),
        t.Key("chunk_overlap_percentage", optional=True, default=None): t.Or(t.Int, t.Null),
        t.Key("chunks_count"): t.Int,
        t.Key("custom_chunking", default=False): t.Bool,
        t.Key("separators", optional=True, default=None): t.Or(
            t.List(t.String(allow_blank=True)), t.Null
        ),
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
        t.Key("organization_id"): t.String,
        t.Key("tenant_id"): t.String,
        t.Key("last_update_date"): t.String,
        t.Key("execution_status"): t.String,
        t.Key("playgrounds_count"): t.Int,
        t.Key("dataset_name"): t.String(allow_blank=True),
        t.Key("user_name"): t.String(allow_blank=True),
        t.Key("source"): t.String,
        t.Key("validation_id", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("error_message", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("embedding_validation_id", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("is_separator_regex"): t.Bool,
        t.Key("parent_id", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("family_id", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("metadata_columns", optional=True, default=None): t.Or(t.Null, t.List(t.String)),
        t.Key("added_dataset_ids", optional=True, default=None): t.Or(t.Null, t.List(t.String)),
        t.Key("added_dataset_names", optional=True, default=None): t.Or(t.Null, t.List(t.String)),
        t.Key("version", optional=True, default=None): t.Or(t.Null, t.Int),
    }
).ignore_extra("*")


class ChunkingParameters(APIObject):
    """
    Parameters defining how documents are split and embedded.

    Attributes
    ----------
    embedding_model : Optional[str]
        Name of the text embedding model.
        Currently supported options are listed in VectorDatabaseEmbeddingModel
        but the values can differ with different platform versions.
    chunking_method : str
        Name of the method to split dataset documents.
        Currently supported options are listed in VectorDatabaseChunkingMethod
        but the values can differ with different platform versions.
    chunk_size : int
        Size of each text chunk in number of tokens.
    chunk_overlap_percentage : int
        Overlap percentage between chunks.
    separators : list[str]
        Strings used to split documents into text chunks.
    embedding_validation : Optional[CustomModelEmbeddingValidation, SupportedCustomModelEmbedding, str]
        ID or object for custom embedding validation.
    custom_chunking : bool
        Determines if the chunking is custom. With custom chunking,
        dataset rows are not split into chunks automatically;
        instead, the user provides the chunks.
    """

    _converter = chunking_parameters_trafaret

    def __init__(
        self,
        embedding_model: Optional[str],
        chunking_method: Union[str, None],
        chunk_size: Union[int, None],
        chunk_overlap_percentage: Union[int, None],
        separators: Union[List[str], None],
        custom_chunking: bool = False,
        embedding_validation: Optional[
            Union[CustomModelEmbeddingValidation, SupportedCustomModelEmbedding, str]
        ] = None,
    ):
        self.embedding_model = embedding_model
        self.chunking_method = chunking_method
        self.chunk_size = chunk_size
        self.chunk_overlap_percentage = chunk_overlap_percentage
        self.separators = separators
        self.embedding_validation_id = self._get_embedding_validation_id(embedding_validation)
        self.custom_chunking = custom_chunking

    @staticmethod
    def _get_embedding_validation_id(
        embedding_validation: Optional[
            Union[CustomModelEmbeddingValidation, SupportedCustomModelEmbedding, str]
        ] = None,
    ) -> Union[str, None]:
        """Get ID of custom embedding validation from supported objects"""
        if isinstance(
            embedding_validation, (CustomModelEmbeddingValidation, SupportedCustomModelEmbedding)
        ):
            return embedding_validation.id
        return embedding_validation


class EmbeddingModel(APIObject):
    """
    A single model for embedding text.

    Attributes
    ----------
    embedding_model : str
        Name of the text embedding model.
        Currently supported options are listed in VectorDatabaseEmbeddingModel
        but the values can differ with different platform versions.
    description : str
        Description of the embedding model.
    max_sequence_length : int
        The model's maximum number of processable input tokens.
    languages : str
        Languages supported by the model.
        Currently supported options are listed in VectorDatabaseDatasetLanguages
        but the values can differ with different platform versions.
    """

    _converter = embedding_model_trafaret

    def __init__(
        self,
        embedding_model: str,
        description: str,
        max_sequence_length: int,
        languages: List[str],
    ):
        self.embedding_model = embedding_model
        self.description = description
        self.max_sequence_length = max_sequence_length
        self.languages = languages


class SupportedCustomModelEmbedding(APIObject):
    """
    All supported custom embedding models for the use case.

    Attributes
    ----------
    id : str
        ID of the custom model embedding validation.
    name : str
        The name of the custom model embedding validation.
    """

    _converter = supported_custom_model_embedding_trafaret

    def __init__(
        self,
        id: str,
        name: str,
    ):
        self.id = id
        self.name = name


class SupportedEmbeddings(APIObject):
    """
    All supported embedding models including the recommended default model.

    Attributes
    ----------
    embedding_models : list[EmbeddingModel]
        All supported embedding models.
    default_embedding_model : str
        Name of the default recommended text embedding model.
        Currently supported options are listed in VectorDatabaseEmbeddingModel
        but the values can differ with different platform versions.
    custom_model_embedding_validations : List[str]
        External embedding models that have been validated
    """

    _converter = supported_embeddings_trafaret

    def __init__(
        self,
        embedding_models: List[Dict[str, Any]],
        default_embedding_model: str,
        custom_model_embedding_validations: List[Dict[str, Any]],
    ):
        self.embedding_models = [
            EmbeddingModel.from_server_data(model) for model in embedding_models
        ]
        self.default_embedding_model = default_embedding_model
        self.custom_model_embedding_validations = [
            SupportedCustomModelEmbedding.from_server_data(validation)
            for validation in custom_model_embedding_validations
        ]


class TextChunkingParameterFields(APIObject):
    """
    Text chunking parameter fields.

    Attributes
    ----------
    name : str
        Parameter name.
    type : str
        Parameter type.
        Currently supported options are listed in VectorDatabaseChunkingParameterType
        but the values can differ with different platform versions.
    description : str
        Parameter description.
    default : int or list[str]
        Parameter default value.
    min : Optional[int]
        Parameter minimum value.
    max : Optional[int]
        Parameter maximum value.
    """

    _converter = text_chunking_parameter_fields_trafaret

    def __init__(
        self,
        name: str,
        type: str,
        description: str,
        default: Union[int, List[str]],
        min: Optional[int] = None,
        max: Optional[int] = None,
    ):
        self.name = name
        self.type = type
        self.description = description
        self.default = default
        self.min = min
        self.max = max


class TextChunkingMethod(APIObject):
    """
    A single text chunking method.

    Attributes
    ----------
    chunking_method : str
        Name of the method to split dataset documents.
        Currently supported options are listed in VectorDatabaseChunkingMethod
        but the values can differ with different platform versions.
    chunking_parameters : list[TextChunkingParameterFields]
        All chunking parameters including ranges and defaults.
    description : str
        Description of the chunking method.
    """

    _converter = text_chunking_method_trafaret

    def __init__(
        self,
        chunking_method: Union[str, None],
        chunking_parameters: List[Dict[str, Any]],
        description: str,
    ):
        self.chunking_method = chunking_method
        self.chunking_parameters = [
            TextChunkingParameterFields.from_server_data(parameter)
            for parameter in chunking_parameters
        ]
        self.description = description


class TextChunkingConfig(APIObject):
    """
    A single text chunking configurations.

    Attributes
    ----------
    embedding_model : str
        Name of the text embedding model.
        Currently supported options are listed in VectorDatabaseEmbeddingModel
        but the values can differ with different platform versions.
    methods : list[TextChunkingMethod]
        All text chunking methods and their supported parameters.
    default_method : str
        The default text chunking method name.
        Currently supported options are listed in VectorDatabaseChunkingMethod
        but the values can differ with different platform versions.
    """

    _converter = text_chunking_config_trafaret

    def __init__(
        self,
        embedding_model: str,
        methods: List[Dict[str, Any]],
        default_method: str,
    ):
        self.embedding_model = embedding_model
        self.methods = [TextChunkingMethod.from_server_data(method) for method in methods]
        self.default_method = default_method


class SupportedTextChunkings(APIObject):
    """
    Supported text chunking configurations which includes a set of
    recommended chunking parameters for each supported embedding model.

    Attributes
    ----------
    text_chunking_configs
        All supported text chunking configurations.
    """

    _converter = supported_text_chunkings_trafaret

    def __init__(self, text_chunking_configs: List[Dict[str, Any]]):
        self.text_chunking_configs = [
            TextChunkingConfig.from_server_data(config) for config in text_chunking_configs
        ]


class SupportedRetrievalSetting(APIObject):
    """
    A single supported retrieval setting.

    Attributes
    ----------
    name: str
        The name of the setting.
    type: str or list[str]
        The type of the setting.
    description: str
        The description of the setting.
    title: str
        The title of the setting.
    default: str, int, bool, or None
        The default value of the setting.
    minimum: int or None
        The minimum value of the setting.
    maximum: int or None
        The maximum value of the setting.
    enum: list[str] or None
        The enum values of the setting.
    settings: list[SupportedRetrievalSetting] or None
        The supported retriever settings.
    group_id: str or None
        The group ID of the setting.
    """

    _converter = supported_retrieval_setting_trafaret

    def __init__(
        self,
        name: str,
        type: Union[str, List[str]],
        description: str,
        title: str,
        default: Union[str, int, bool, None],
        minimum: Union[int, None],
        maximum: Union[int, None],
        enum: Union[List[str], None],
        settings: Union[List[Union[Dict[str, Any], List[Dict[str, Any]]]], None],
        group_id: Union[str, None],
    ):
        self.name = name
        self.type = type
        self.description = description
        self.title = title
        self.default = default
        self.minimum = minimum
        self.maximum = maximum
        self.enum = enum
        self.settings = (
            [SupportedRetrievalSetting.from_server_data(setting) for setting in settings]
            if settings
            else None
        )
        self.group_id = group_id


class SupportedRetrievalSettings(APIObject):
    """
    Supported retrieval settings.

    Attributes
    ----------
    settings : list[dict]
        The supported retriever settings.
    """

    _converter = supported_retrieval_settings_trafaret

    def __init__(
        self,
        settings: List[Union[Dict[str, Any], List[Dict[str, Any]]]],
    ):
        self.settings = [
            SupportedRetrievalSetting.from_server_data(setting) for setting in settings
        ]


class VectorDatabaseDatasetExportJob(APIObject):
    """
    Response for the vector database dataset export job.

    Attributes
    ----------
    job_id : str
        ID of the export job.
    vector_database_id : str
        ID of the vector database.
    export_dataset_id : str
        ID of the exported dataset.
    """

    _converter = vector_database_dataset_export_job_trafaret

    def __init__(
        self,
        job_id: str,
        vector_database_id: str,
        export_dataset_id: str,
    ):
        self.job_id = job_id
        self.vector_database_id = vector_database_id
        self.export_dataset_id = export_dataset_id


class VectorDatabase(APIObject):
    """
    Metadata for a DataRobot vector database accessible to the user.

    Attributes
    ----------
    id : str
        Vector database ID.
    name : str
        Vector database name.
    size : int
        Size of the vector database assets in bytes.
    use_case_id : str
        Linked use case ID.
    dataset_id : str
        ID of the dataset used for creation.
    embedding_model : str
        Name of the text embedding model.
        Currently supported options are listed in VectorDatabaseEmbeddingModel
        but the values can differ with different platform versions.
    chunking_method : str or None
        Name of the method to split dataset documents.
        Currently supported options are listed in VectorDatabaseChunkingMethod
        but the values can differ with different platform versions.
    chunk_size : int or None
        Size of each text chunk in number of tokens.
    chunk_overlap_percentage : int or None
        Overlap percentage between chunks.
    chunks_count : int
        Total number of text chunks.
    custom_chunking : bool
        Determines if the chunking is custom. With custom chunking,
        dataset rows are not split into chunks automatically;
        instead, the user provides the chunks.
    separators : list[string] or None
        Separators for document splitting.
    creation_date : str
        Date when the database was created.
    creation_user_id : str
        The ID of the creating user.
    organization_id : str
        The creating user's organization ID.
    tenant_id : str
        The creating user's tenant ID.
    last_update_date : str
        Last update date for the database.
    execution_status : str
        Database execution status.
        Currently supported options are listed in VectorDatabaseExecutionStatus
        but the values can differ with different platform versions.
    playgrounds_count : int
        Number of using playgrounds.
    dataset_name : str
        Name of the used dataset.
    user_name : str
        Name of the creating user.
    source : str
        Source of the vector database.
        Currently supported options are listed in VectorDatabaseSource
        but the values can differ with different platform versions.
    validation_id : Optional[str]
        ID of custom model vector database validation.
        Only filled for external vector databases.
    error_message : Optional[str]
        Additional information for errored vector database.
    embedding_validation_id : Optional[str]
        ID of the custom embedding validation, if any.
    is_separator_regex : bool
        Whether the separators should be treated as regular expressions.
    """

    _path = "api/v2/genai/vectorDatabases"

    _converter = vector_database_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        size: int,
        use_case_id: str,
        dataset_id: Optional[str],
        embedding_model: Optional[str],
        chunking_method: Union[Optional[str], None],
        chunk_size: Union[int, None],
        chunk_overlap_percentage: Union[int, None],
        chunks_count: int,
        custom_chunking: bool,
        separators: Union[List[str], None],
        creation_date: str,
        creation_user_id: str,
        organization_id: str,
        tenant_id: str,
        last_update_date: str,
        execution_status: str,
        playgrounds_count: int,
        dataset_name: str,
        user_name: str,
        source: str,
        validation_id: Optional[str],
        error_message: Optional[str],
        is_separator_regex: bool,
        embedding_validation_id: Optional[str],
        # VDB_PERFORMANCE fields
        parent_id: Optional[str],
        family_id: Optional[str],
        metadata_columns: Optional[List[str]],
        added_dataset_ids: Optional[List[str]],
        added_dataset_names: Optional[List[str]],
        version: int,
    ):
        self.id = id
        self.name = name
        self.size = size
        self.use_case_id = use_case_id
        self.dataset_id = dataset_id
        self.embedding_model = embedding_model
        self.chunking_method = chunking_method
        self.chunk_size = chunk_size
        self.chunk_overlap_percentage = chunk_overlap_percentage
        self.chunks_count = chunks_count
        self.separators = separators
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.organization_id = organization_id
        self.tenant_id = tenant_id
        self.last_update_date = last_update_date
        self.execution_status = execution_status
        self.playgrounds_count = playgrounds_count
        self.dataset_name = dataset_name
        self.user_name = user_name
        self.source = source
        self.validation_id = validation_id
        self.error_message = error_message
        self.is_separator_regex = is_separator_regex
        self.embedding_validation_id = embedding_validation_id
        self.custom_chunking = custom_chunking
        self.parent_id = parent_id
        self.family_id = family_id
        self.metadata_columns = metadata_columns
        self.added_dataset_ids = added_dataset_ids
        self.added_dataset_names = added_dataset_names
        self.version = version

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.id}, name={self.name}, "
            f"execution_status={self.execution_status})"
        )

    @classmethod
    def get_supported_embeddings(
        cls,
        dataset_id: Optional[str] = None,
        use_case: Optional[Union[UseCase, str]] = None,
    ) -> SupportedEmbeddings:
        """Get all supported and the recommended embedding models.

        Parameters
        ----------
        dataset_id : Optional[str]
            ID of a dataset for which the recommended model is returned
            based on the detected language of that dataset.
        use_case : Optional[UseCase, str]
            May be Use Case ID or the Use Case entity.

        Returns
        -------
        supported_embeddings : SupportedEmbeddings
            The supported embedding models.
        """
        params = {
            "dataset_id": dataset_id,
            "use_case_id": get_use_case_id(use_case, is_required=False),
        }
        url = f"{cls._client.domain}/{cls._path}/supportedEmbeddings/"
        r_data = cls._client.get(url, params=params)
        return SupportedEmbeddings.from_server_data(r_data.json())

    def submit_export_dataset_job(
        self,
    ) -> VectorDatabaseDatasetExportJob:
        """Submit the vector database dataset export job.

        Returns
        -------
        result : VectorDatabaseDatasetExportJob
            The result of the vector database dataset export job containing the exported dataset id.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/datasetExportJobs/"
        r_data = self._client.post(url)
        return VectorDatabaseDatasetExportJob.from_server_data(r_data.json())

    @classmethod
    def get_supported_retrieval_settings(cls) -> SupportedRetrievalSettings:
        """Get supported retrieval settings.

        Returns
        -------
        supported_retrieval_settings : SupportedRetrievalSettings
            The supported retriever settings.
        """
        url = f"{cls._client.domain}/{cls._path}/supportedRetrievalSettings/"
        r_data = cls._client.get(url)
        return SupportedRetrievalSettings.from_server_data(r_data.json())

    @classmethod
    def create(
        cls,
        dataset_id: str,
        chunking_parameters: Optional[ChunkingParameters] = None,
        use_case: Optional[Union[UseCase, str]] = None,
        name: Optional[str] = None,
        parent_vector_database_id: Optional[str] = None,
        update_llm_blueprints: Optional[bool] = None,
        update_deployments: Optional[bool] = None,
    ) -> VectorDatabase:
        """
        Create a new vector database.

        Parameters
        ----------
        dataset_id : str
            ID of the dataset used for creation.
        chunking_parameters : ChunkingParameters
            Parameters defining how documents are split and embedded.
        use_case : Optional[Union[UseCase, str]], optional
            Use case to link to the created vector database.
        name : Optional[str]
            Vector database name, by default None
            which leads to the default name 'Vector Database for <dataset name>'.
        parent_vector_database_id : Optional[str]
            ID of the parent vector database to base the update on.
        update_llm_blueprints : Optional[bool]
            Whether to update LLM blueprints related to the parent vector database.
        update_deployments : Optional[bool]
            Whether to update deployments related to the parent vector database.

        Returns
        -------
        vector database : VectorDatabase
            The created vector database with execution status 'new'.
        """
        payload = {
            "name": name,
            "dataset_id": dataset_id,
            "use_case_id": get_use_case_id(use_case, is_required=True),
            "chunking_parameters": chunking_parameters,
            "parent_vector_database_id": parent_vector_database_id,
            "update_llm_blueprints": update_llm_blueprints,
            "update_deployments": update_deployments,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, data=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def create_from_custom_model(
        cls,
        name: str,
        use_case: Optional[Union[UseCase, str]] = None,
        validation_id: Optional[str] = None,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        deployment_id: Optional[str] = None,
        model_id: Optional[str] = None,
    ) -> VectorDatabase:
        """
        Create a new vector database from validated custom model deployment.

        Parameters
        ----------
        name : str
            Vector database name.
        use_case : Optional[Union[UseCase, str]], optional
            Use case to link to the created vector database.
        validation_id : Optional[str]
            ID of CustomModelVectorDatabaseValidation for the deployment.
            Alternatively, you can specify ALL the following fields.
        prompt_column_name : Optional[str]
            The column name the deployed model expect as the input.
        target_column_name : Optional[str]
            The target name deployed model will output.
        deployment_id : Optional[str]
            ID of the deployment.
        model_id : Optional[str]
            ID of the underlying deployment model.
            Can be found from the API as Deployment.model["id"].

        Returns
        -------
        vector database : VectorDatabase
            The created vector database.
        """
        payload: Dict[str, Any] = {
            "use_case_id": get_use_case_id(use_case, is_required=True),
            "name": name,
        }
        if validation_id:
            payload["validation_id"] = validation_id
        else:
            assert all([prompt_column_name, target_column_name, deployment_id, model_id]), (
                "Either validation_id or all of prompt_column_name, target_column_name, "
                "deployment_id, model_id fields should be specified"
            )
            payload.update(
                {
                    "prompt_column_name": prompt_column_name,
                    "target_column_name": target_column_name,
                    "deployment_id": deployment_id,
                    "model_id": model_id,
                }
            )
        url = f"{cls._client.domain}/{cls._path}/fromCustomModelDeployment/"
        r_data = cls._client.post(url, data=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def get(cls, vector_database_id: str) -> VectorDatabase:
        """
        Retrieve a single vector database.

        Parameters
        ----------
        vector_database_id : str
            The ID of the vector database you want to retrieve.

        Returns
        -------
        vector database : VectorDatabase
            The requested vector database.
        """
        url = f"{cls._client.domain}/{cls._path}/{vector_database_id}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        use_case: Optional[UseCaseLike] = None,
        playground: Optional[Union[Playground, str]] = None,
        search: Optional[str] = None,
        sort: Optional[str] = None,
        completed_only: Optional[bool] = None,
    ) -> List[VectorDatabase]:
        """
        List all vector databases associated with a specific use case available to the user.

        Parameters
        ----------
        use_case : Optional[UseCaseLike], optional
            The returned vector databases are filtered to those associated with a specific Use Case
            or Cases if specified or can be inferred from the Context.
            Accepts either the entity or the ID.
        playground : Optional[Union[Playground, str]], optional
            The returned vector databases are filtered to those associated with a specific playground
            if it is specified. Accepts either the entity or the ID.
        search : Optional[str]
            String for filtering vector databases.
            Vector databases that contain the string in name will be returned.
            If not specified, all vector databases will be returned.
        sort : Optional[str]
            Property to sort vector databases by.
            Prefix the attribute name with a dash to sort in descending order,
            e.g. sort='-creationDate'.
            Currently supported options are listed in ListVectorDatabasesSortQueryParams
            but the values can differ with different platform versions.
            By default, the sort parameter is None which will result in
            vector databases being returned in order of creation time descending.
        completed_only : Optional[bool]
            A filter to retrieve only vector databases that have been successfully created.
            By default, all vector databases regardless of execution status are retrieved.

        Returns
        -------
        vectorbases : list[VectorDatabase]
            A list of vector databases available to the user.
        """
        params = {
            "search": search,
            "sort": sort,
            "completed_only": completed_only,
            "playground_id": get_entity_id(playground) if playground else None,
        }

        params = resolve_use_cases(use_cases=use_case, params=params, use_case_key="use_case_id")

        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def update(self, name: str) -> VectorDatabase:
        """
        Update the vector database.

        Parameters
        ----------
        name : str
            The new name for the vector database.

        Returns
        -------
        vector database : VectorDatabase
            The updated vector database.
        """
        payload = {"name": name}
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        r_data = self._client.patch(url, data=payload)
        return self.from_server_data(r_data.json())

    def delete(self) -> None:
        """
        Delete the vector database.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)

    @classmethod
    def get_supported_text_chunkings(cls) -> SupportedTextChunkings:
        """Get all supported text chunking configurations which includes
        a set of recommended chunking parameters for each supported embedding model.

        Returns
        -------
        supported_text_chunkings : SupportedTextChunkings
            The supported text chunking configurations.
        """
        url = f"{cls._client.domain}/{cls._path}/supportedTextChunkings/"
        r_data = cls._client.get(url)
        return SupportedTextChunkings.from_server_data(r_data.json())

    def download_text_and_embeddings_asset(self, file_path: Optional[str] = None) -> None:
        """Download a parquet file with text chunks and corresponding embeddings created
        by a vector database.

        Parameters
        ----------
        file_path : Optional[str]
            File path to save the asset. By default, it saves in the current directory
            autogenerated by server name.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/textAndEmbeddings/"
        response = self._client.get(url, stream=True)
        if not file_path:
            file_path = response.headers["Content-Disposition"].split("=")[-1].strip('"')
        with open(str(file_path), mode="wb") as f:
            for chunk in response.iter_content(chunk_size=65536):
                f.write(chunk)


class CustomModelVectorDatabaseValidation(NonChatAwareCustomModelValidation):
    """
    Validation record checking the ability of the deployment to serve as a vector database.

    Attributes
    ----------
    id : str
        The ID of the validation.
    prompt_column_name : str
        The column name the deployed model expect as the input.
    target_column_name : str
        The target name deployed model will output.
    deployment_id : str
        ID of the deployment.
    model_id : str
        ID of the underlying deployment model.
        Can be found from the API as Deployment.model["id"].
    validation_status : str
        Can be TESTING, FAILED and PASSED. Only PASSED allowed for use.
    deployment_access_data : dict, optional
        The data that will be used for accessing the deployment prediction server.
        This field is only available for deployments that pass validation.
        Dict fields are as follows:
        - prediction_api_url - The URL for the deployment prediction server.
        - datarobot_key - The first of two auth headers for the prediction server.
        - authorization_header - The second of two auth headers for the prediction server.
        - input_type - The input type the model expects, either JSON or CSV.
        - model_type - The target type of the deployed custom model.
    tenant_id : str
        The creating user's tenant ID.
    name : str
        The display name of the validated custom model.
    creation_date : str
        The creation date of the validation (ISO 8601 formatted).
    user_id : str
        The ID of the creating user.
    error_message : Optional[str]
        Additional information for the errored validation.
    deployment_name : Optional[str]
        The name of the validated deployment.
    user_name : Optional[str]
        The name of the creating user.
    use_case_id : Optional[str]
        The ID of the Use Case associated with the validation.
    """

    _path = "api/v2/genai/customModelVectorDatabaseValidations"



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from mypy_extensions import TypedDict
import trafaret as t

from datarobot.enums import (
    AggregationType,
    enum_to_list,
    GuardConditionComparator,
    GuardType,
    InsightTypes,
)
from datarobot.models.api_object import APIObject
from datarobot.models.genai.comparison_chat import get_entity_id
from datarobot.models.genai.playground import Playground
from datarobot.models.genai.playground_moderation_configuration import (
    moderation_configuration_with_id,
    moderation_configuration_without_id,
)
from datarobot.models.use_cases.use_case import UseCase
from datarobot.models.use_cases.utils import get_use_case_id


class InsightsConfigurationDict(TypedDict):
    """Dictionary representation of insights configuration."""

    insight_name: str
    insight_type: Optional[str]
    deployment_id: Optional[str]
    model_id: Optional[str]
    sidecar_model_metric_validation_id: Optional[str]
    custom_metric_id: Optional[str]
    evaluation_dataset_configuration_id: Optional[str]
    cost_configuration_id: Optional[str]
    result_unit: Optional[str]
    ootb_metric_id: Optional[str]
    ootb_metric_name: Optional[str]
    guard_conditions: Optional[List[Dict[str, Any]]]
    moderation_configuration: Optional[Dict[str, Any]]
    execution_status: Optional[str]
    error_message: Optional[str]
    error_resolution: Optional[str]
    nemo_metric_id: Optional[str]
    llm_id: Optional[str]
    custom_model_llm_validation_id: Optional[str]
    stage: Optional[str]
    aggregation_types: Optional[List[str]]
    sidecar_model_metric_metadata: Optional[Dict[str, Any]]
    guard_template_id: Optional[str]
    guard_configuration_id: Optional[str]
    model_package_registered_model_id: Optional[str]


guard_conditions = t.Dict(
    {
        t.Key("comparator"): t.Enum(*enum_to_list(GuardConditionComparator)),
        t.Key("comparand"): t.Or(t.String, t.Float, t.Bool, t.List(t.String)),
    }
).ignore_extra("*")

sidecar_model_metric_metadata = t.Dict(
    {
        t.Key("prompt_column_name", optional=True): t.Or(t.String, t.Null),
        t.Key("response_column_name", optional=True): t.Or(t.String, t.Null),
        t.Key("target_column_name", optional=True): t.Or(t.String, t.Null),
        t.Key("expected_response_column_name", optional=True): t.Or(t.String, t.Null),
        t.Key("guard_type", optional=True): t.Or(t.Enum(*enum_to_list(GuardType)), t.Null),
    }
).ignore_extra("*")

insight_configuration_trafaret = t.Dict(
    {
        t.Key("insight_name"): t.String,
        t.Key("insight_type", optional=True): t.Or(t.Enum(*enum_to_list(InsightTypes)), t.Null),
        t.Key("deployment_id", optional=True): t.Or(t.String, t.Null),
        t.Key("model_id", optional=True): t.Or(t.String, t.Null),
        t.Key("sidecar_model_metric_validation_id", optional=True): t.Or(t.String, t.Null),
        t.Key("custom_metric_id", optional=True): t.Or(t.String, t.Null),
        t.Key("evaluation_dataset_configuration_id", optional=True): t.Or(t.String, t.Null),
        t.Key("cost_configuration_id", optional=True): t.Or(t.String, t.Null),
        t.Key("result_unit", optional=True): t.Or(t.String, t.Null),
        t.Key("ootb_metric_id", optional=True): t.Or(t.String, t.Null),
        t.Key("ootb_metric_name", optional=True): t.Or(t.String, t.Null),
        t.Key("guard_conditions", optional=True): t.Or(t.List(guard_conditions), t.Null),
        t.Key("moderation_configuration", optional=True): t.Or(
            moderation_configuration_with_id, moderation_configuration_without_id, t.Null
        ),
        t.Key("execution_status", optional=True): t.Or(t.String, t.Null),
        t.Key("error_message", optional=True): t.Or(t.String, t.Null),
        t.Key("error_resolution", optional=True): t.Or(t.String, t.Null),
        t.Key("nemo_metric_id", optional=True): t.Or(t.String, t.Null),
        t.Key("llm_id", optional=True): t.Or(t.String, t.Null),
        t.Key("custom_model_llm_validation_id", optional=True): t.Or(t.String, t.Null),
        t.Key("stage", optional=True): t.Or(t.String, t.Null),
        # additional data
        t.Key("aggregation_types", optional=True): t.Or(
            t.List(t.Enum(*enum_to_list(AggregationType))), t.Null
        ),
        t.Key("sidecar_model_metric_metadata", optional=True): t.Or(
            sidecar_model_metric_metadata, t.Null
        ),
        t.Key("guard_template_id", optional=True): t.Or(t.String, t.Null),
        t.Key("guard_configuration_id", optional=True): t.Or(t.String, t.Null),
        t.Key("model_package_registered_model_id", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")

insights_trafaret = t.Dict(
    {
        t.Key("playground_id"): t.String,
        t.Key("insights_configuration"): t.List(insight_configuration_trafaret),
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
        t.Key("last_update_date"): t.String,
        t.Key("last_update_user_id"): t.String,
        t.Key("tenant_id"): t.String,
    }
).ignore_extra("*")


class InsightsConfiguration(APIObject):
    """
    Configuration information for a specific insight.

    Attributes
    ----------
    insight_name : str
        The name of the insight.
    insight_type : InsightTypes, optional
        The type of the insight.
    deployment_id : Optional[str]
        The deployment ID the insight is applied to.
    model_id : Optional[str]
        The model ID for the insight.
    sidecar_model_metric_validation_id : Optional[str]
        Validation ID for the sidecar model metric.
    custom_metric_id : Optional[str]
        The ID for a custom model metric.
    evaluation_dataset_configuration_id : Optional[str]
        The ID for the evaluation dataset configuration.
    cost_configuration_id : Optional[str]
        The ID for the cost configuration information.
    result_unit : Optional[str]
        The unit of the result, for example "USD".
    ootb_metric_id : Optional[str]
        The ID of the Datarobot-provided metric that does not require additional configuration.
    ootb_metric_name : Optional[str]
        The name of the Datarobot-provided metric that does not require additional configuration.
    guard_conditions : list[dict], optional
        The guard conditions to be used with the insight.
    moderation_configuration : dict, optional
        The moderation configuration for the insight.
    execution_status : Optional[str]
        The execution status of the insight.
    error_message : Optional[str]
        The error message for the insight, for example if it is missing specific configuration
        for deployed models.
    error_resolution : Optional[str]
        An indicator of which field must be edited to resolve an error state.
    nemo_metric_id : Optional[str]
        The ID for the NEMO metric.
    llm_id : Optional[str]
        The LLM ID for OOTB metrics that use LLMs.
    custom_model_llm_validation_id : Optional[str]
        The ID for the custom model LLM validation if using a custom model LLM for OOTB metrics.
    aggregation_types : list[str], optional
        The aggregation types to be used for the insight.
    stage : Optional[str]
        The stage (prompt or response) when the metric is calculated.
    sidecar_model_metric_metadata : dict, optional
        Metadata specific to sidecar model metrics.
    guard_template_id : Optional[str]
        The ID for the guard template that applies to the insight.
    guard_configuration_id : Optional[str]
        The ID for the guard configuration that applies to the insight.
    """

    _converter = insight_configuration_trafaret

    def __init__(
        self,
        insight_name: str,
        insight_type: Optional[str] = None,
        deployment_id: Optional[str] = None,
        model_id: Optional[str] = None,
        sidecar_model_metric_validation_id: Optional[str] = None,
        custom_metric_id: Optional[str] = None,
        evaluation_dataset_configuration_id: Optional[str] = None,
        cost_configuration_id: Optional[str] = None,
        result_unit: Optional[str] = None,
        ootb_metric_id: Optional[str] = None,
        ootb_metric_name: Optional[str] = None,
        guard_conditions: Optional[List[Dict[str, Any]]] = None,
        moderation_configuration: Optional[Dict[str, Any]] = None,
        execution_status: Optional[str] = None,
        error_message: Optional[str] = None,
        error_resolution: Optional[str] = None,
        nemo_metric_id: Optional[str] = None,
        llm_id: Optional[str] = None,
        custom_model_llm_validation_id: Optional[str] = None,
        stage: Optional[str] = None,
        aggregation_types: Optional[List[str]] = None,
        sidecar_model_metric_metadata: Optional[Dict[str, Any]] = None,
        guard_template_id: Optional[str] = None,
        guard_configuration_id: Optional[str] = None,
        model_package_registered_model_id: Optional[str] = None,
    ):
        self.insight_name = insight_name
        self.insight_type = insight_type
        self.deployment_id = deployment_id
        self.model_id = model_id
        self.sidecar_model_metric_validation_id = sidecar_model_metric_validation_id
        self.custom_metric_id = custom_metric_id
        self.evaluation_dataset_configuration_id = evaluation_dataset_configuration_id
        self.cost_configuration_id = cost_configuration_id
        self.result_unit = result_unit
        self.ootb_metric_id = ootb_metric_id
        self.ootb_metric_name = ootb_metric_name
        self.guard_conditions = guard_conditions
        self.moderation_configuration = moderation_configuration
        self.execution_status = execution_status
        self.error_message = error_message
        self.error_resolution = error_resolution
        self.nemo_metric_id = nemo_metric_id
        self.llm_id = llm_id
        self.custom_model_llm_validation_id = custom_model_llm_validation_id
        self.stage = stage
        self.aggregation_types = aggregation_types
        self.sidecar_model_metric_metadata = sidecar_model_metric_metadata
        self.guard_template_id = guard_template_id
        self.guard_configuration_id = guard_configuration_id
        self.model_package_registered_model_id = model_package_registered_model_id

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.insight_name}, model_id={self.model_id})"

    def to_dict(self) -> InsightsConfigurationDict:
        return InsightsConfigurationDict(
            insight_name=self.insight_name,
            insight_type=self.insight_type,
            deployment_id=self.deployment_id,
            model_id=self.model_id,
            sidecar_model_metric_validation_id=self.sidecar_model_metric_validation_id,
            custom_metric_id=self.custom_metric_id,
            evaluation_dataset_configuration_id=self.evaluation_dataset_configuration_id,
            cost_configuration_id=self.cost_configuration_id,
            result_unit=self.result_unit,
            ootb_metric_id=self.ootb_metric_id,
            ootb_metric_name=self.ootb_metric_name,
            guard_conditions=self.guard_conditions,
            moderation_configuration=self.moderation_configuration,
            execution_status=self.execution_status,
            error_message=self.error_message,
            error_resolution=self.error_resolution,
            nemo_metric_id=self.nemo_metric_id,
            llm_id=self.llm_id,
            custom_model_llm_validation_id=self.custom_model_llm_validation_id,
            stage=self.stage,
            aggregation_types=self.aggregation_types,
            sidecar_model_metric_metadata=self.sidecar_model_metric_metadata,
            guard_template_id=self.guard_template_id,
            guard_configuration_id=self.guard_configuration_id,
            model_package_registered_model_id=self.model_package_registered_model_id,
        )


class SupportedInsights(InsightsConfiguration):
    """Supported insights configurations for a given use case."""

    _path = "api/v2/genai/insights/supportedInsights/"

    @classmethod
    def list(cls, use_case_id: str) -> List[InsightsConfiguration]:
        """Get a list of all supported insights that can be used within a given Use Case.

        Parameters
        ----------
        use_case_id: str
            The ID of the Use Case to list supported insights for.

        Returns
        -------
        insights: list[InsightsConfiguration]
            A list of supported insights.
        """
        response_data = cls._client.get(
            url=f"{cls._client.domain}/{cls._path}", params={"use_case_id": use_case_id}
        )
        return [
            cls.from_server_data(insight)
            for insight in response_data.json()["insightsConfiguration"]
        ]


class Insights(APIObject):
    """
    The insights configured for a playground.

    Attributes
    ----------
    playground_id : str
        The ID of the playground the insights are configured for.
    insights_configuration : list[InsightsConfiguration]
        The insights configuration for the playground.
    creation_date : str
        The date the insights were configured.
    creation_user_id : str
        The ID of the user who created the insights.
    last_update_date : str
        The date the insights were last updated.
    last_update_user_id : str
        The ID of the user who last updated the insights.
    tenant_id : str
        The tenant ID that applies to the record.

    """

    _path = "api/v2/genai/insights"

    _converter = insights_trafaret

    def __init__(
        self,
        playground_id: str,
        insights_configuration: List[Dict[str, Any]],
        creation_date: str,
        creation_user_id: str,
        last_update_date: str,
        last_update_user_id: str,
        tenant_id: str,
    ):
        self.playground_id = playground_id
        self.insights_configuration = [
            InsightsConfiguration.from_server_data(config) for config in insights_configuration
        ]
        self.creation_date = creation_date
        self.creation_user_Id = creation_user_id
        self.last_update_date = last_update_date
        self.last_update_user_id = last_update_user_id
        self.tenant_id = tenant_id

    @classmethod
    def get(
        cls, playground: Union[str, Playground], with_aggregation_types_only: bool = False
    ) -> Insights:
        """Get the insights configuration for a given playground.

        Parameters
        ----------
        playground: str|Playground
            The ID of the playground to get insights for.
        with_aggregation_types_only: Optional[bool]
            If True, only return the aggregation types for the insights.

        Returns
        -------
        insights: Insights
            The insights configuration for the playground.
        """
        playground_id = get_entity_id(playground)
        response_data = cls._client.get(
            url=f"{cls._client.domain}/{cls._path}/{playground_id}/",
            params={"with_aggreation_types_only": with_aggregation_types_only},
        )
        return cls.from_server_data(response_data.json())

    @classmethod
    def create(
        cls,
        playground: Union[str, Playground],
        insights_configuration: List[InsightsConfiguration],
        use_case: Union[UseCase, str],
    ) -> Insights:
        """Create a new insights configuration for a given playground.

        Parameters
        ----------
        playground: str
            The ID of the playground to create insights for.
        insights_configuration: list[InsightsConfiguration]
            The insights configuration for the playground.
        use_case_id: str
            The Use Case ID to the playground is a part of.

        Returns
        -------
        insights: Insights
            The created insights configuration.
        """
        playground_id = get_entity_id(playground)
        payload = {
            "playground_id": playground_id,
            "insights_configuration": [config.to_dict() for config in insights_configuration],
            "use_case_id": get_use_case_id(use_case, is_required=True),
        }

        response_data = cls._client.post(url=f"{cls._client.domain}/{cls._path}/", json=payload)
        return cls.from_server_data(response_data.json())



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, cast, Dict, List, Optional

import trafaret as t
from typing_extensions import Self

from datarobot.enums import INSIGHTS_SOURCES
from datarobot.insights.base import BaseInsight


class ShapPreview(BaseInsight):
    """Class for SHAP Preview calculations. Use the standard methods of BaseInsight to compute
    and retrieve: compute, create, list, get.
    """

    SHAP_PREVIEW_VALUE = t.Dict(
        {
            t.Key("feature_rank"): t.Int(),
            t.Key("feature_name"): t.String(),
            t.Key("feature_value"): t.String(),
            t.Key("shap_value"): t.Float(),
        }
    ).ignore_extra("*")

    SHAP_PREVIEW_ROW = t.Dict(
        {
            t.Key("row_index"): t.Int(),
            t.Key("total_preview_features"): t.Int(),
            t.Key("prediction_value"): t.Float(),
            t.Key("preview_values"): t.List(SHAP_PREVIEW_VALUE),
        }
    )

    INSIGHT_NAME = "shapPreview"
    INSIGHT_DATA = {
        t.Key("previews_count"): t.Int(),
        t.Key("previews"): t.List(SHAP_PREVIEW_ROW),
    }

    @property
    def previews(self) -> List[Dict[str, Any]]:
        """SHAP preview values.

        Returns
        -------
        preview : List[Dict[str, Any]]
            A list of the ShapPreview values for each row.
        """
        return cast(List[Dict[str, Any]], self.data["previews"])

    @property
    def previews_count(self) -> int:
        """The number of shap preview rows.

        Returns
        -------
        int
        """
        return cast(int, self.data["previews_count"])

    @classmethod
    def get(
        cls,
        entity_id: str,
        source: str = INSIGHTS_SOURCES.VALIDATION,
        quick_compute: Optional[bool] = None,
        prediction_filter_row_count: Optional[int] = None,
        prediction_filter_percentiles: Optional[int] = None,
        prediction_filter_operand_first: Optional[float] = None,
        prediction_filter_operand_second: Optional[float] = None,
        prediction_filter_operator: Optional[str] = None,
        feature_filter_count: Optional[int] = None,
        feature_filter_name: Optional[str] = None,
        **kwargs: Any,
    ) -> Self:
        """Return the first matching ShapPreview insight based on the entity id and kwargs.

        Parameters
        ----------
        entity_id: str
            The ID of the entity to retrieve generated insights.
        source: str
            The source type to use when retrieving the insight.
        quick_compute: Optional[bool]
            Sets whether to retrieve the insight that was computed using quick-compute. If not
            specified, quick_compute is not used for matching.
        prediction_filter_row_count: Optional[int]
            The maximum number of preview rows to return.
        prediction_filter_percentiles: Optional[int]
            The number of percentile intervals to select from the total number of rows.
            This field will supersede predictionFilterRowCount if both are present.
        prediction_filter_operand_first: Optional[float]
            The first operand to apply to filtered predictions.
        prediction_filter_operand_second: Optional[float]
            The second operand to apply to filtered predictions.
        prediction_filter_operator: Optional[str]
            The operator to apply to filtered predictions.
        feature_filter_count: Optional[int]
            The maximum number of features to return for each preview.
        feature_filter_name: Optional[str]
            The names of specific features to return for each preview.

        Returns
        -------
        List[Any]
            List of newly or already computed insights.
        """
        query_params = {**kwargs}
        if prediction_filter_row_count is not None:
            query_params["predictionFilterRowCount"] = prediction_filter_row_count
        if prediction_filter_percentiles is not None:
            query_params["predictionFilterPercentiles"] = prediction_filter_percentiles
        if prediction_filter_operand_first is not None:
            query_params["predictionFilterOperandFirst"] = prediction_filter_operand_first
        if prediction_filter_operand_second is not None:
            query_params["predictionFilterOperandSecond"] = prediction_filter_operand_second
        if prediction_filter_operator is not None:
            query_params["predictionFilterOperator"] = prediction_filter_operator
        if feature_filter_count is not None:
            query_params["featureFilterCount"] = feature_filter_count
        if feature_filter_name is not None:
            query_params["featureFilterName"] = feature_filter_name

        return super().get(
            entity_id=entity_id, source=source, quick_compute=quick_compute, **query_params
        )



#
# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from collections import namedtuple
from typing import cast, Dict, List, Optional, TYPE_CHECKING, Union

import trafaret as t
from typing_extensions import NotRequired

from datarobot.enums import DEFAULT_MAX_WAIT, PROJECT_STAGE
from datarobot.helpers.image_utils import get_image_from_bytes
from datarobot.models.api_object import APIObject
from datarobot.models.project import Project
from datarobot.utils import from_api
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution

__all__ = [
    "DocumentPageFile",
    "DocumentThumbnail",
    "DocumentTextExtractionSamplePage",
    "DocumentTextExtractionSample",
]

FeaturesWithSamples = namedtuple(
    "FeaturesWithSamples", ["model_id", "feature_name", "document_task"]
)


if TYPE_CHECKING:
    from PIL.Image import Image
    from mypy_extensions import TypedDict

    class PredictionType(TypedDict):
        values: Union[float, List[float]]
        labels: NotRequired[List[str]]

    class TextLine(TypedDict):
        left: int
        top: int
        right: int
        bottom: int
        text: str


class DocumentPageFile(APIObject):
    """Page of a document as an image file.

    Attributes
    ----------
    project_id : str
        The identifier of the project which the document page belongs to.
    document_page_id : str
        The unique identifier for the document page.
    height : int
         The height of the document thumbnail in pixels.
    width : int
         The width of the document thumbnail in pixels.
    thumbnail_bytes : bytes
        The number of bytes of the document thumbnail image. Accessing this may
        require a server request and an associated delay in fetching the resource.
    mime_type : str
        The mime image type of the document thumbnail. Example: `'image/png'`
    """

    _bytes_path = "projects/{project_id}/documentPages/{document_page_id}/file/"

    def __init__(
        self,
        document_page_id: str,
        project_id: Optional[str] = None,
        height: int = 0,
        width: int = 0,
        download_link: Optional[str] = None,
    ):
        self.project_id = project_id
        self.document_page_id = document_page_id
        self.height = height
        self.width = width
        self.download_link = (
            download_link
            if download_link
            else self._bytes_path.format(project_id=project_id, document_page_id=document_page_id)
        )
        self.__thumbnail_bytes: bytes
        self.__mime_type: str

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(project_id={self.project_id}, document_page_id={self.document_page_id}, "
            f"height={self.height}, width={self.width})"
        )

    @property
    def thumbnail_bytes(self) -> bytes:
        """Document thumbnail as bytes.

        Returns
        -------
        bytes
            Document thumbnail.
        """
        if not getattr(self, "__thumbnail_bytes", None):
            self.__get_thumbnail_bytes()
        return self.__thumbnail_bytes

    @property
    def mime_type(self) -> str:
        """Mime image type of the document thumbnail. Example: `'image/png'`

        Returns
        -------
        str
            Mime image type of the document thumbnail.
        """
        if not getattr(self, "__mime_type", None):
            # Getting and setting thumbnail bytes also gets and sets mime type
            self.__get_thumbnail_bytes()
        return self.__mime_type

    def __get_thumbnail_bytes(self) -> None:
        """Method that fetches document thumbnail from the server and
         sets the `mime_type` and `thumbnail_bytes` properties.

        Returns
        -------
        None
        """
        r_data = self._client.get(self.download_link)
        self.__mime_type = r_data.headers.get("Content-Type")
        self.__thumbnail_bytes = r_data.content


class DocumentThumbnail(APIObject):
    """Thumbnail of document from the project's dataset.

    If ``Project.stage`` is ``datarobot.enums.PROJECT_STAGE.EDA2``
    and it is a supervised project then the ``target_*`` attributes
    of this class will have values, otherwise the values will all be None.

    Attributes
    ----------
    document: Document
        The document object.
    project_id : str
        The identifier of the project which the document thumbnail belongs to.
    target_value: str
        The target value used for filtering thumbnails.
    """

    _list_eda_sample_path = "projects/{project_id}/documentThumbnailSamples/"
    _list_project_sample_path = "projects/{project_id}/documentThumbnails/"

    _converter = t.Dict(
        {
            t.Key("project_id"): t.String(),
            t.Key("document_page_id"): t.String(),
            t.Key("height"): t.Int(),
            t.Key("width"): t.Int(),
            t.Key("target_value", optional=True): t.Or(
                t.String(), t.Int(), t.Float(), t.List(t.String())
            ),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id: str,
        document_page_id: str,
        height: int = 0,
        width: int = 0,
        target_value: Optional[Union[str, int, float, List[str]]] = None,
    ):
        self.document = DocumentPageFile(
            project_id=project_id, document_page_id=document_page_id, height=height, width=width
        )
        self.project_id = project_id
        self.target_value = target_value

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(project_id={self.project_id}, "
            f"document_page_id={self.document.document_page_id}, target_value={self.target_value})"
        )

    @classmethod
    def list(
        cls,
        project_id: str,
        feature_name: str,
        target_value: Optional[str] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[DocumentThumbnail]:
        """Get document thumbnails from a project.

        Parameters
        ----------
        project_id : str
            The identifier of the project which the document thumbnail belongs to.
        feature_name : str
            The name of feature that specifies the document type.
        target_value : Optional[str], default ``None``
            The target value to filter thumbnails.
        offset : Optional[int], default ``None``
            The number of documents to be skipped.
        limit : Optional[int], default ``None``
            The number of document thumbnails to return.

        Returns
        -------
        documents : List[DocumentThumbnail]
            A list of ``DocumentThumbnail`` objects, each representing a single document.

        Notes
        -----
        Actual document thumbnails are not fetched from the server by this method.
        Instead the data gets loaded lazily when ``DocumentPageFile`` object attributes
        are accessed.

        Examples
        --------

        Fetch document thumbnails for the given ``project_id`` and ``feature_name``.

        .. code-block:: python

            from datarobot._experimental.models.documentai.document import DocumentThumbnail

            # Fetch five documents from the EDA SAMPLE for the specified project and specific feature
            document_thumbs = DocumentThumbnail.list(project_id, feature_name, limit=5)

            # Fetch five documents for the specified project with target value filtering
            # This option is only available after selecting the project target and starting modeling
            target1_thumbs = DocumentThumbnail.list(project_id, feature_name, target_value='target1', limit=5)


        Preview the document thumbnail.

        .. code-block:: python

            from datarobot._experimental.models.documentai.document import DocumentThumbnail
            from datarobot.helpers.image_utils import get_image_from_bytes

            # Fetch 3 documents
            document_thumbs = DocumentThumbnail.list(project_id, feature_name, limit=3)

            for doc_thumb in document_thumbs:
                thumbnail = get_image_from_bytes(doc_thumb.document.thumbnail_bytes)
                thumbnail.show()
        """
        project = Project.get(project_id=project_id)

        if project.stage in [PROJECT_STAGE.EDA2, PROJECT_STAGE.MODELING]:
            path = cls._list_project_sample_path.format(project_id=project_id)
        else:
            path = cls._list_eda_sample_path.format(project_id=project_id)

        list_params = dict(featureName=feature_name)
        if target_value:
            list_params["targetValue"] = target_value
        if offset:
            list_params["offset"] = str(offset)
        if limit:
            list_params["limit"] = str(limit)

        r_data = cls._client.get(path, params=list_params).json()
        documents = []

        # construct document objects for each document sample
        for doc_data in r_data["data"]:
            doc_data["project_id"] = project_id
            document: DocumentThumbnail = cls.from_server_data(doc_data)
            documents.append(document)
        return documents


COMMON_CONVERTER_DICT_KEYS = {
    t.Key("document_index"): t.Int(),
    t.Key("feature_name"): t.String(),
    t.Key("actual_target_value", optional=True): t.Or(
        t.Null(), t.String(), t.Int(), t.Float(), t.List(t.String)
    ),
    t.Key("prediction", optional=True): t.Or(
        t.Null(),
        t.Dict(
            {
                t.Key("values"): t.Or(t.Float(), t.List(t.Float())),
                t.Key("labels", optional=True): t.Or(t.Null(), t.List(t.String())),
            }
        ),
    ),
    t.Key("document_task"): t.String(),
}


class DocumentTextExtractionSampleDocument(APIObject):
    """Document text extraction source.

    Holds data that contains feature and model prediction values, as well as the thumbnail of the document.

    Attributes
    ----------
    document_index: int
        The index of the document page sample.
    feature_name: str
        The name of the feature that the document text extraction sample is related to.
    thumbnail_id: str
        The document page ID.
    thumbnail_width: int
        The thumbnail image width.
    thumbnail_height: int
        The thumbnail image height.
    thumbnail_link: str
        The thumbnail image download link.
    document_task: str
        The document blueprint task that the document belongs to.
    actual_target_value: Optional[Union[str, int, List[str]]]
        The actual target value.
    prediction: Optional[PredictionType]
        Prediction values and labels.
    """

    _documents_path = "models/{model_id}/documentTextExtractionSampleDocuments/"

    _converter_dict_keys = {
        t.Key("thumbnail_id"): t.String(),
        t.Key("thumbnail_width"): t.Int(),
        t.Key("thumbnail_height"): t.Int(),
        t.Key("thumbnail_link"): t.String(),
    }
    _converter_dict_keys.update(COMMON_CONVERTER_DICT_KEYS)
    _converter = t.Dict(_converter_dict_keys).ignore_extra("*")

    def __init__(
        self,
        document_index: int,
        feature_name: str,
        thumbnail_id: str,
        thumbnail_width: int,
        thumbnail_height: int,
        thumbnail_link: str,
        document_task: str,
        actual_target_value: Optional[Union[str, int, float, List[str]]] = None,
        prediction: Optional[PredictionType] = None,
    ) -> None:
        self.document_index = document_index
        self.feature_name = feature_name
        self.actual_target_value = actual_target_value
        self.prediction = prediction
        self.thumbnail = DocumentPageFile(
            project_id=None,
            document_page_id=thumbnail_id,
            height=thumbnail_height,
            width=thumbnail_width,
            download_link=thumbnail_link,
        )
        self.document_task = document_task

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(document_index={self.document_index}, "
            f"feature_name={self.feature_name}, actual_target_value={self.actual_target_value}, "
            f"document_task={self.document_task})"
        )

    @classmethod
    def list(
        cls, model_id: str, feature_name: str, document_task: Optional[str] = None
    ) -> List[DocumentTextExtractionSampleDocument]:
        """
        List available documents with document text extraction samples.

        Parameters
        ----------
        model_id: str
            The identifier for the model.
        feature_name: str
            The name of the feature,
        document_task: Optional[str]
            The document blueprint task.

        Returns
        -------
        List[DocumentTextExtractionSampleDocument]
        """
        list_documents_path = cls._documents_path.format(model_id=model_id)
        params = dict(feature_name=feature_name)
        if document_task is not None:
            params["document_task"] = document_task
        return [
            cls.from_server_data(page)
            for page in unpaginate(list_documents_path, params, cls._client)
        ]


class DocumentTextExtractionSamplePage(APIObject):
    """Document text extraction sample covering one document page.

    Holds data about the document page, the recognized text, and the location of the text in the document page.

    Attributes
    ----------
    page_index: int
        Index of the page inside the document
    document_index: int
        Index of the document inside the dataset
    feature_name: str
        The name of the feature that the document text extraction sample belongs to.
    document_page_id: str
        The document page ID.
    document_page_width: int
        Document page width.
    document_page_height: int
        Document page height.
    document_page_link: str
        Document page link to download the document page image.
    text_lines: List[Dict[str, Union[int, str]]]
        A list of text lines and their coordinates.
    document_task: str
        The document blueprint task that the page belongs to.
    actual_target_value: Optional[Union[str, int, List[str]]
        Actual target value.
    prediction: Optional[PredictionType]
        Prediction values and labels.
    """

    _pages_path = "models/{model_id}/documentTextExtractionSamplePages/"
    _converter_dict_keys = {
        t.Key("page_index"): t.Int(),
        t.Key("document_index"): t.Int(),
        t.Key("document_page_id"): t.String(),
        t.Key(
            "document_page_width",
        ): t.Int(),
        t.Key("document_page_height"): t.Int(),
        t.Key("document_page_link"): t.String(),
        t.Key(
            "text_lines",
        ): t.List(
            t.Dict(
                {
                    t.Key("left"): t.Int(),
                    t.Key("top"): t.Int(),
                    t.Key("right"): t.Int(),
                    t.Key("bottom"): t.Int(),
                    t.Key("text"): t.String(),
                }
            )
        ),
    }
    _converter_dict_keys.update(COMMON_CONVERTER_DICT_KEYS)
    _converter = t.Dict(_converter_dict_keys).ignore_extra("*")

    def __init__(
        self,
        page_index: int,
        document_index: int,
        feature_name: str,
        document_page_id: str,
        document_page_width: int,
        document_page_height: int,
        document_page_link: str,
        text_lines: List[TextLine],
        document_task: str,
        actual_target_value: Optional[Union[str, int, float, List[str]]] = None,
        prediction: Optional[PredictionType] = None,
    ) -> None:
        self.page_index = page_index
        self.document_page = DocumentTextExtractionSampleDocument(
            document_index=document_index,
            feature_name=feature_name,
            actual_target_value=actual_target_value,
            prediction=prediction,
            thumbnail_id=document_page_id,
            thumbnail_height=document_page_height,
            thumbnail_width=document_page_width,
            thumbnail_link=document_page_link,
            document_task=document_task,
        )
        self.text_lines = text_lines

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(page_index={self.page_index}, "
            f"text_lines_count={len(self.text_lines)}, document_page={repr(self.document_page)})"
        )

    @classmethod
    def list(
        cls,
        model_id: str,
        feature_name: str,
        document_index: Optional[int] = None,
        document_task: Optional[str] = None,
    ) -> List[DocumentTextExtractionSamplePage]:
        """
        Returns a list of document text extraction sample pages.

        Parameters
        ----------
        model_id: str
            The model identifier, used to retrieve document text extraction page samples.
        feature_name: str
            The feature name, used to retrieve document text extraction page samples.
        document_index: Optional[int]
            The specific document index to retrieve. Defaults to None.
        document_task: Optional[str]
            Document blueprint task.

        Returns
        -------
        List[DocumentTextExtractionSamplePage]
        """
        list_pages_path = cls._pages_path.format(model_id=model_id)
        params = dict(feature_name=feature_name)
        if document_index is not None:
            params["document_index"] = str(document_index)
        if document_task is not None:
            params["document_task"] = document_task
        return [cls.from_server_data(x) for x in unpaginate(list_pages_path, params, cls._client)]

    def get_document_page_with_text_locations(
        self,
        line_color: str = "blue",
        line_width: int = 3,
        padding: int = 3,
    ) -> Image:
        """
        Returns the document page with bounding boxes drawn around the text lines as a PIL.Image.

        Parameters
        ----------
        line_color: str
            The color used to draw a bounding box on the image page. Defaults to blue.
        line_width: int
            The line width of the bounding boxes that will be drawn. Defaults to 3.
        padding: int
            The additional space left between the text and the bounding box, measured in pixels. Defaults to 3.

        Returns
        -------
        Image
            Returns a PIL.Image with drawn text-bounding boxes.
        """
        try:
            from PIL import ImageDraw  # pylint: disable=import-outside-toplevel
        except ImportError:
            msg = (
                "Image transformation operations require installation of datarobot library, "
                "with optional `images` dependency. To install library with image support"
                "please use `pip install datarobot[images]`"
            )
            raise ImportError(msg)

        image = get_image_from_bytes(self.document_page.thumbnail.thumbnail_bytes)

        # for each text object draw bounding box on original thumbnail
        image_draw = ImageDraw.Draw(image)
        for bbox in self.text_lines:
            top = bbox["top"] - padding
            bottom = bbox["bottom"] + padding
            left = bbox["left"] - padding
            right = bbox["right"] + padding
            shape = [left, top, right, bottom]
            image_draw.rectangle(xy=shape, outline=line_color, width=line_width)
        return image


class DocumentTextExtractionSample(APIObject):
    """Stateless class for computing and retrieving Document Text Extraction Samples.

    Notes
    -----
    Actual document text extraction samples are not fetched from the server in the moment of
    a function call. Detailed information on the documents, the pages and the rendered images of them
    are fetched when accessed on demand (lazy loading).

    Examples
    --------

    1) Compute text extraction samples for a specific model, and fetch all existing document text
    extraction samples for a specific project.

    .. code-block:: python

        from datarobot._experimental.models.documentai.document import DocumentTextExtractionSample

        SPECIFIC_MODEL_ID1 = "model_id1"
        SPECIFIC_MODEL_ID2 = "model_id2"
        SPECIFIC_PROJECT_ID = "project_id"

        # Order computation of document text extraction sample for specific model.
        # By default `compute` method will await for computation to end before returning
        DocumentTextExtractionSample.compute(SPECIFIC_MODEL_ID1, await_completion=False)
        DocumentTextExtractionSample.compute(SPECIFIC_MODEL_ID2)

        samples = DocumentTextExtractionSample.list_features_with_samples(SPECIFIC_PROJECT_ID)


    2) Fetch document text extraction samples for a specific `model_id` and `feature_name`, and
    display all document sample pages.

    .. code-block:: python

        from datarobot._experimental.models.documentai.document import DocumentTextExtractionSample
        from datarobot.helpers.image_utils import get_image_from_bytes

        SPECIFIC_MODEL_ID = "model_id"
        SPECIFIC_FEATURE_NAME = "feature_name"

        samples = DocumentTextExtractionSample.list_pages(
            model_id=SPECIFIC_MODEL_ID,
            feature_name=SPECIFIC_FEATURE_NAME
        )
        for sample in samples:
            thumbnail = sample.document_page.thumbnail
            image = get_image_from_bytes(thumbnail.thumbnail_bytes)
            image.show()


    3) Fetch document text extraction samples for specific `model_id` and `feature_name` and
    display text extraction details for the first page. This example displays the image of the document
    with bounding boxes of detected text lines. It also returns a list of all text
    lines extracted from page along with their coordinates.

    .. code-block:: python

        from datarobot._experimental.models.documentai.document import DocumentTextExtractionSample

        SPECIFIC_MODEL_ID = "model_id"
        SPECIFIC_FEATURE_NAME = "feature_name"

        samples = DocumentTextExtractionSample.list_pages(SPECIFIC_MODEL_ID, SPECIFIC_FEATURE_NAME)
        # Draw bounding boxes for first document page sample and display related text data.
        image = samples[0].get_document_page_with_text_locations()
        image.show()
        # For each text block represented as bounding box object drawn on original image
        # display its coordinates (top, left, bottom, right) and extracted text value
        for text_line in samples[0].text_lines:
            print(text_line)

    """

    _text_extraction_samples_compute = "models/{model_id}/documentTextExtractionSamples/"
    _text_extraction_computed_samples_list = "projects/{project_id}/documentTextExtractionSamples/"

    @classmethod
    def compute(
        cls, model_id: str, await_completion: bool = True, max_wait: int = DEFAULT_MAX_WAIT
    ) -> None:
        """
        Starts computation of document text extraction samples for the model and, if successful,
        returns computed text samples for it. This method allows calculation to continue for
        a specified time and, if not complete, cancels the request.

        Parameters
        ----------
        model_id: str
            The identifier of the project's model that start the creation of the cluster insights.
        await_completion: bool
            Determines whether the method should wait for completion before exiting or not.
        max_wait: int (default=600)
            The maximum number of seconds to wait for the request to finish before raising an
            AsyncTimeoutError.

        Raises
        ------
        ClientError
            Server rejected creation due to client error.
            Often, a bad `model_id` is causing these errors.
        AsyncFailureError
            Indicates whether any of the responses from the server are unexpected.
        AsyncProcessUnsuccessfulError
            Indicates whether the cluster insights computation failed or was cancelled.
        AsyncTimeoutError
            Indicates whether the cluster insights computation did not resolve within the specified
            time limit (`max_wait`).
        """
        compute_path = cls._text_extraction_samples_compute.format(model_id=model_id)
        response = cls._client.post(compute_path)
        if await_completion:
            wait_for_async_resolution(cls._client, response.headers["Location"], max_wait=max_wait)

    @classmethod
    def list_features_with_samples(cls, project_id: str) -> List[FeaturesWithSamples]:
        """
        Returns a list of features, `model_id` pairs with computed document text extraction samples.

        Parameters
        ----------
        project_id: str
            The project ID to retrieve the list of computed samples for.

        Returns
        -------
        List[FeaturesWithSamples]
        """
        list_samples_path = cls._text_extraction_computed_samples_list.format(project_id=project_id)
        return [
            FeaturesWithSamples(**cast(Dict[str, str], from_api(x)))
            for x in unpaginate(list_samples_path, {}, cls._client)
        ]

    @classmethod
    def list_pages(
        cls,
        model_id: str,
        feature_name: str,
        document_index: Optional[int] = None,
        document_task: Optional[str] = None,
    ) -> List[DocumentTextExtractionSamplePage]:
        """
        Returns a list of document text extraction sample pages.

        Parameters
        ----------
        model_id: str
            The model identifier.
        feature_name: str
            The specific feature name to retrieve.
        document_index: Optional[int]
            The specific document index to retrieve. Defaults to None.
        document_task: Optional[str]
            The document blueprint task.

        Returns
        -------
        List[DocumentTextExtractionSamplePage]
        """
        return DocumentTextExtractionSamplePage.list(
            model_id=model_id,
            feature_name=feature_name,
            document_index=document_index,
            document_task=document_task,
        )

    @classmethod
    def list_documents(
        cls, model_id: str, feature_name: str
    ) -> List[DocumentTextExtractionSampleDocument]:
        """
        Returns a list of documents used for text extraction.

        Parameters
        ----------
        model_id: str
            The model identifier.
        feature_name: str
            The feature name.

        Returns
        -------
        List[DocumentTextExtractionSampleDocument]
        """
        return DocumentTextExtractionSampleDocument.list(
            model_id=model_id, feature_name=feature_name
        )



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from io import StringIO
from typing import Any, cast, Dict, Iterable, List, Optional, Union

import pandas as pd
import trafaret as t
from typing_extensions import Self

from datarobot.enums import DEFAULT_MAX_WAIT, ENTITY_TYPES, INSIGHTS_SOURCES
from datarobot.mixins.browser_mixin import BrowserMixin
from datarobot.mixins.datarobot_ui_mixin import DatarobotUIMixin
from datarobot.models import StatusCheckJob
from datarobot.models.api_object import APIObject
from datarobot.utils import pagination

DATAROBOT_INSIGHTS_FEATURE_IMPACT_MIMETYPE = "application/vnd.datarobot.insight.feature-impact"


class BaseInsight(APIObject, BrowserMixin, DatarobotUIMixin):
    """Base Insight class for modern insights

    This class serves as a template for modern insights created using the Root Insights framework.
    It provides most necessary functions for easily implementing classes that wrap specific insights.

    """

    # To use this class it should be inherited and the `INSIGHT_NAME` and `INSIGHT_DATA` attributes
    # should be replaced with the correct values.
    #
    # Example:
    # You have an insight that has the endpoint `insights/myInsight` and the insight's calculator
    # produces a response that contains a boolean field called `insight_worked` and a list of
    # string values called `insight_values`.
    #
    # To make things easier for the user you may want to add some easily accessible properties
    # to allow them to quickly get the values of the insight.
    #
    # Example insight result:
    #
    # .. code-block:: text
    #
    #         {
    #             "insight_worked": true,
    #             "insight_values": ["hello", "world"]
    #         }
    #
    # You can simply create a wrapper for this new insight using this framework:
    #
    # .. code-block:: python
    #
    #         class MyInsight(BaseInsight):
    #             INSIGHT_NAME = "myInsight"
    #             INSIGHT_DATA = {
    #                 t.Key("insight_worked"): t.Bool(),
    #                 t.Key("insight_values"): t.List(t.String()),
    #             }
    #
    #             @property
    #             def values() -> List[str]:
    #                 return cast(List[str], self.data["insight_values"])
    #
    # All of the computation and insight retrieval handling will be accounted for by the base insight.
    # This class can also be overridden for more complex queries by adjusting the functionality of the
    # `_get_payload` function to provide additional required options, but most insights won't need
    # this capability.

    INSIGHT_NAME: str = "base"  # Defined by subclasses to identify the formal name for endpoints
    INSIGHT_DATA = t.Any()  # Defined by subclasses for specific insights

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("entity_id"): t.String(),
            t.Key("project_id"): t.String(),
            t.Key("source"): t.String(),
            t.Key("data_slice_id", optional=True): t.Or(t.String(), t.Null()),
            t.Key("external_dataset_id", optional=True): t.Or(t.String(), t.Null()),
            t.Key("data"): INSIGHT_DATA,
        }
    ).ignore_extra("*")

    # all feature impact types are rendered with the same UI components mapped with this mime type
    _datarobot_ui_mime_type = DATAROBOT_INSIGHTS_FEATURE_IMPACT_MIMETYPE

    def __init__(
        self,
        id: str,
        entity_id: str,
        project_id: str,
        source: str,
        data: Any,
        data_slice_id: Optional[str] = None,
        external_dataset_id: Optional[str] = None,
    ):
        """Base Insight class

        Parameters
        ----------
        id: str
            ID of the insight.
        entity_id: str
            ID of the entity associated with the insight.
        project_id: str
            ID of the project associated with the insight.
        source: str
            Source type of the insight.
        data: Dict
            Results of the computed insight.
        data_slice_id: Optional[str]
            Data slice ID associated with the insight.
        external_dataset_id: Optional[str]
            External dataset ID associated with the insight.
        """
        self.id = id
        self.entity_id = entity_id
        self.project_id = project_id
        self.source = source
        self.data_slice_id = data_slice_id
        self.external_dataset_id = external_dataset_id

        self.data = data

    @classmethod
    def _compute_url(cls) -> str:
        """Construct a compute url"""
        return f"insights/{cls.INSIGHT_NAME}/"

    @classmethod
    def _retrieve_url(cls) -> str:
        """Construct a retrieve url"""
        return f"insights/{cls.INSIGHT_NAME}" + "/models/{}/"

    def get_uri(self) -> str:
        """This should define the URI to their browser based interactions"""
        raise NotImplementedError

    def sort(self, key_name: str) -> None:
        """Sorts insights data"""
        raise NotImplementedError

    @classmethod
    # type: ignore[override]
    def from_server_data(
        cls,
        data: Dict[str, Any],
        keep_attrs: Optional[Iterable[str]] = None,
    ) -> Self:
        """Override from_server_data to handle paginated responses"""
        if all(attrib in data for attrib in ["count", "next", "previous"]):
            # If true, the data response is a paginated response from a compute command and the data
            # entity should be unwrapped. There should only be one response.
            data = data["data"][0]

        return super().from_server_data(data=data, keep_attrs=keep_attrs)

    @classmethod
    def _get_payload(
        cls,
        entity_id: str,
        source: str = INSIGHTS_SOURCES.VALIDATION,
        data_slice_id: Optional[str] = None,
        external_dataset_id: Optional[str] = None,
        entity_type: Optional[ENTITY_TYPES] = None,
        quick_compute: Optional[bool] = None,
        **kwargs: Any,
    ) -> Dict[str, Union[str, int, bool]]:
        """Construct a payload for a compute request. May be overridden by insight subclasses to
        accept additional parameters.
        """
        _ = kwargs
        payload: Dict[str, Union[str, bool]] = {
            "entityId": entity_id,
            "source": source,
        }
        if data_slice_id is not None:
            payload["dataSliceId"] = data_slice_id
        if external_dataset_id is not None:
            payload["externalDatasetId"] = external_dataset_id
        if entity_type is not None:
            payload["entityType"] = entity_type
        if quick_compute is not None:
            payload["quickCompute"] = quick_compute

        return cast(Dict[str, Union[str, int, bool]], payload)

    @classmethod
    def compute(
        cls,
        entity_id: str,
        source: str = INSIGHTS_SOURCES.VALIDATION,
        data_slice_id: Optional[str] = None,
        external_dataset_id: Optional[str] = None,
        entity_type: Optional[ENTITY_TYPES] = ENTITY_TYPES.DATAROBOT_MODEL,
        quick_compute: Optional[bool] = None,
        **kwargs: Any,
    ) -> StatusCheckJob:
        """Submit an insight compute request. You can use `create` if you want to
        wait synchronously for the completion of the job. May be overridden by insight subclasses to
        accept additional parameters.

        Parameters
        ----------
        entity_id: str
            The ID of the entity to compute the insight.
        source: str
            The source type to use when computing the insight.
        data_slice_id: Optional[str]
            Data slice ID to use when computing the insight.
        external_dataset_id: Optional[str]
            External dataset ID to use when computing the insight.
        entity_type: Optional[ENTITY_TYPES]
            The type of the entity associated with the insight. Select one of the ENTITY_TYPE enum
            values, or accept the default, "datarobotModel".
        quick_compute: Optional[bool]
            Sets whether to use quick-compute for the insight. If `True` or unspecified, the insight
            is computed using a 2500-row data sample. If `False`, the insight is computed using all
            rows in the chosen source.

        Returns
        -------
        StatusCheckJob
            Status check job entity for the asynchronous insight calculation.
        """
        payload = cls._get_payload(
            entity_id=entity_id,
            source=source,
            data_slice_id=data_slice_id,
            external_dataset_id=external_dataset_id,
            entity_type=entity_type,
            quick_compute=quick_compute,
            **kwargs,
        )
        response = cls._client.post(cls._compute_url(), data=payload)

        return StatusCheckJob.from_response(response, cls)

    @classmethod
    def create(
        cls,
        entity_id: str,
        source: str = INSIGHTS_SOURCES.VALIDATION,
        data_slice_id: Optional[str] = None,
        external_dataset_id: Optional[str] = None,
        entity_type: Optional[ENTITY_TYPES] = ENTITY_TYPES.DATAROBOT_MODEL,
        quick_compute: Optional[bool] = None,
        max_wait: Optional[int] = DEFAULT_MAX_WAIT,
        **kwargs: Any,
    ) -> Self:
        """Create an insight and wait for completion. May be overridden by insight subclasses to
        accept additional parameters.

        Parameters
        ----------
        entity_id: str
            The ID of the entity to compute the insight.
        source: str
            The source type to use when computing the insight.
        data_slice_id: Optional[str]
            Data slice ID to use when computing the insight.
        external_dataset_id: Optional[str]
            External dataset ID to use when computing the insight.
        entity_type: Optional[ENTITY_TYPES]
            The type of the entity associated with the insight. Select one of the ENTITY_TYPE enum
            values, or accept the default, "datarobotModel".
        quick_compute: Optional[bool]
            Sets whether to use quick-compute for the insight. If `True` or unspecified, the insight
            is computed using a 2500-row data sample. If `False`, the insight is computed using all
            rows in the chosen source.
        max_wait: int
            The number of seconds to wait for the result.

        Returns
        -------
        Self
            Entity of the newly or already computed insights.
        """
        status_check_job = cls.compute(
            entity_id=entity_id,
            source=source,
            data_slice_id=data_slice_id,
            external_dataset_id=external_dataset_id,
            entity_type=entity_type,
            quick_compute=quick_compute,
            **kwargs,
        )
        return cast(Self, status_check_job.get_result_when_complete(max_wait=cast(int, max_wait)))

    @classmethod
    def list(cls, entity_id: str) -> List[Self]:
        """List all generated insights.

        Parameters
        ----------
        entity_id: str
            The ID of the entity queried for listing all generated insights.

        Returns
        -------
        List[Self]
            List of newly or previously computed insights.
        """
        url = cls._retrieve_url().format(entity_id)
        return [cls.from_server_data(x) for x in pagination.unpaginate(url, {}, cls._client)]

    @classmethod
    def get(
        cls,
        entity_id: str,
        source: str = INSIGHTS_SOURCES.VALIDATION,
        quick_compute: Optional[bool] = None,
        **kwargs: Any,
    ) -> Self:
        """Return the first matching insight based on the entity id and kwargs.

        Parameters
        ----------
        entity_id: str
            The ID of the entity to retrieve generated insights.
        source: str
            The source type to use when retrieving the insight.
        quick_compute: Optional[bool]
            Sets whether to retrieve the insight that was computed using quick-compute. If not
            specified, quick_compute is not used for matching.

        Returns
        -------
        Self
            Previously computed insight.
        """
        url = cls._retrieve_url().format(entity_id)
        query_params: Dict[str, Union[str, bool]] = {"source": source, **kwargs}
        if quick_compute is not None:
            query_params["quickCompute"] = quick_compute
        return [
            cls.from_server_data(x) for x in pagination.unpaginate(url, query_params, cls._client)
        ][0]


class CsvSupportedInsight(BaseInsight):
    """Base Insight class for insights that support CSV export

    This class serves as a template for insights that support CSV export. It provides additional
    functionality to export the insight results to a CSV file.

    """

    CSV_HEADERS = {"Content-Type": "text/csv", "Accept": "text/csv"}

    @classmethod
    def get_as_csv(cls, entity_id: str, **kwargs: Any) -> str:
        """Retrieve a specific insight represented in CSV format.

        Parameters
        ----------
        entity_id: str
            ID of the entity to retrieve the insight.
        **kwargs: Any
            Additional keyword arguments to pass to the retrieve function.

        Returns
        -------
        str
            The retrieved insight.
        """
        url = cls._retrieve_url().format(entity_id)
        resp_data = cls._client.get(url, params=kwargs, headers=cls.CSV_HEADERS)
        return cast(str, resp_data.text)

    @classmethod
    def get_as_dataframe(cls, entity_id: str, **kwargs: Any) -> pd.DataFrame:
        """Retrieve a specific insight represented as a pandas DataFrame.

        Parameters
        ----------
        entity_id: str
            ID of the entity to retrieve the insight.
        **kwargs: Any
            Additional keyword arguments to pass to the retrieve function.

        Returns
        -------
        DataFrame
            The retrieved insight.
        """
        csv_data = cls.get_as_csv(entity_id, **kwargs)
        return pd.read_csv(StringIO(csv_data))



# flake8: noqa
# because the unused imports are on purpose

from .base import BaseInsight, CsvSupportedInsight
from .lift_chart import LiftChart
from .residuals import Residuals
from .roc_curve import RocCurve
from .shap_distributions import ShapDistributions
from .shap_impact import ShapImpact
from .shap_matrix import ShapMatrix
from .shap_preview import ShapPreview



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, cast, Dict, List

import trafaret as t

from datarobot.insights.base import BaseInsight


class ShapDistributions(BaseInsight):
    """Class for SHAP Distributions calculations. Use the standard methods of BaseInsight to compute
    and retrieve: compute, create, list, get.
    """

    SHAP_DISTRIBUTION_VALUE = t.Dict(
        {
            t.Key("row_index"): t.Int(),
            t.Key("prediction_value"): t.Float(),
            t.Key("feature_rank"): t.Int(),
            t.Key("feature_value"): t.String(),
            t.Key("shap_value"): t.Float(),
        }
    ).ignore_extra("*")

    SHAP_DISTRIBUTIONS_ROW = t.Dict(
        {
            t.Key("feature"): t.String(),
            t.Key("shap_values"): t.List(SHAP_DISTRIBUTION_VALUE),
            t.Key("impact_unnormalized"): t.Float(),
            t.Key("impact_normalized"): t.Float(),
            t.Key("feature_type"): t.String(),
        }
    )

    INSIGHT_NAME = "shapDistributions"
    INSIGHT_DATA = {
        t.Key("total_features_count"): t.Int(),
        t.Key("features"): t.List(SHAP_DISTRIBUTIONS_ROW),
    }

    @property
    def features(self) -> List[Dict[str, Any]]:
        """SHAP feature values

        Returns
        -------
        features : List[Dict[str, Any]]
            A list of the ShapDistributions values for each row
        """
        return cast(List[Dict[str, Any]], self.data["features"])

    @property
    def total_features_count(self) -> int:
        """Number of shap distributions features

        Returns
        -------
        int
        """
        return cast(int, self.data["total_features_count"])



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from dataclasses import dataclass
from enum import auto, Enum
from pathlib import Path
from typing import List

import trafaret as t

from datarobot.models.api_object import APIObject, ServerDataType
from datarobot.utils.pagination import unpaginate


class OCRJobDatasetLanguage(Enum):
    """Supported OCR language"""

    ENGLISH = auto()
    JAPANESE = auto()

    @classmethod
    def from_string(cls, language_str: str) -> "OCRJobDatasetLanguage":
        for language in cls:
            if language_str == language.name:
                return language
        msg = (
            f"Language str {language_str} is invalid. "
            f"Valid values are {[lang.name for lang in cls]}"
        )
        raise ValueError(msg)


def from_language_enum_to_api_representation(lang_enum: OCRJobDatasetLanguage) -> str:
    """Convert OCRJobDatasetLanguage to API representation.

    Parameters
    ----------
    lang_enum: OCRJobDatasetLanguage
        language enum

    Returns
    -------
    str
        OCRJobDatasetLanguage API representation
    """
    return lang_enum.name


class OCRJobStatusEnum(Enum):
    """OCR Job status enum"""

    EXECUTING = auto()
    FAILURE = auto()
    PENDING = auto()
    STOPPED = auto()
    SUCCESS = auto()
    UNKNOWN = auto()


def from_api_representation_to_job_status_enum(status_api_representation: str) -> OCRJobStatusEnum:
    """Convert OCR job status API representation to OCRJobStatusEnum.

    Parameters
    ----------
    status_api_representation: str
        OCR job status API representation

    Returns
    -------
    OCRJobStatusEnum
        OCR job status enum

    Raises
    ------
    ValueError
        Raised when status_api_representation is invalid
    """
    if not status_api_representation.islower():
        raise ValueError("Status api representation must be lowercase.")
    for status in list(OCRJobStatusEnum):
        if status_api_representation == status.name.lower():
            return status
    msg = (
        f"Status str {status_api_representation} is invalid. "
        f"Valid values are {[status.name.lower() for status in list(OCRJobStatusEnum)]}"
    )
    raise ValueError(msg)


@dataclass
class StartOCRJobResponse:
    """Container of Start OCR Job API response"""

    job_status_location: str
    output_location: str
    error_report_location: str

    @classmethod
    def from_server_data(cls, data: ServerDataType) -> "StartOCRJobResponse":
        return cls(
            job_status_location=data["jobStatusLocation"],  # type: ignore[call-overload]
            output_location=data["outputLocation"],  # type: ignore[call-overload]
            error_report_location=data["errorReportLocation"],  # type: ignore[call-overload]
        )


class OCRJobResource(APIObject):
    """An OCR job resource container. It is used to:
    - Get an existing OCR  job resource.
    - List available OCR job resources.
    - Start an OCR job.
    - Check the status of a started OCR job.
    - Download the error report of a started OCR job.

    .. versionadded:: v3.6.0b0

    Attributes
    ----------
    id: str
        The identifier of an OCR job resource.
    input_catalog_id: str
        The identifier of an AI catalog item used as the OCR job input.
    output_catalog_id: str
        The identifier of an AI catalog item used as the OCR job output.
    user_id: str
        The identifier of a user.
    job_started: bool
        Determines if a job associated with the OCRJobResource has started.
    language: str
        String representation of OCRJobDatasetLanguage.
    """

    _path = "ocrJobResources/"
    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("input_catalog_id"): t.String(),
            t.Key("output_catalog_id"): t.String(),
            t.Key("user_id"): t.String(),
            t.Key("job_started"): t.Bool(),
            t.Key("language"): t.Enum(*[el.name for el in OCRJobDatasetLanguage]),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        id: str,
        input_catalog_id: str,
        output_catalog_id: str,
        user_id: str,
        job_started: bool,
        language: str,
    ) -> None:
        self.id = id
        self.input_catalog_id = input_catalog_id
        self.output_catalog_id = output_catalog_id
        self.user_id = user_id
        self.job_started = job_started
        self.language = OCRJobDatasetLanguage.from_string(language)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id!r})"

    @classmethod
    def get(cls, job_resource_id: str) -> "OCRJobResource":
        """Get an OCR job resource.

        Parameters
        ----------
        job_resource_id: str
            identifier of OCR job resource

        Returns
        -------
        OCRJobResource
            returned OCR job resource
        """
        path = f"{cls._path}{job_resource_id}/"
        data = cls._client.get(path).json()
        return cls.from_server_data(data)

    @classmethod
    def list(cls, offset: int = 0, limit: int = 10) -> List["OCRJobResource"]:
        """Get a list of OCR job resources.

        Parameters
        ----------
        offset: int
            The offset of the query.
        limit: int
            The limit of returned OCR job resources.

        Returns
        -------
        List[OCRJobResource]
            A list of OCR job resources.
        """
        query_params = {"offset": offset, "limit": limit}
        list_of_job_resources = unpaginate(cls._path, query_params, cls._client)
        return [
            OCRJobResource.from_server_data(job_resource) for job_resource in list_of_job_resources
        ]

    @classmethod
    def create(cls, input_catalog_id: str, language: OCRJobDatasetLanguage) -> "OCRJobResource":
        """Create a new OCR job resource and return it.

        Parameters
        ----------
        input_catalog_id: str
            The identifier of an AI catalog item used as the OCR job input.
        language: OCRJobDatasetLanguage
            The OCR job dataset language.

        Returns
        -------
        OCRJobResource
            The created OCR job resource.
        """
        request_body = {
            "dataset_id": input_catalog_id,
            "language": from_language_enum_to_api_representation(language),
        }
        data = cls._client.post(
            cls._path,
            data=request_body,
            headers={"Content-type": "application/json"},
        ).json()
        return cls.from_server_data(data)

    def start_job(self) -> StartOCRJobResponse:
        """Start an OCR job with this OCR job resource.

        Returns
        -------
        StartOCRJobResponse
            The response of starting an OCR job.
        """
        data = self._client.post(f"{self._path}{self.id}/start/").json()
        return StartOCRJobResponse.from_server_data(data)

    def get_job_status(self) -> OCRJobStatusEnum:
        """Get status of the OCR job associated with this OCR job resource.

        Returns
        -------
        OCRJobStatusEnum
            OCR job status enum
        """
        path = f"{self._path}{self.id}/jobStatus/"
        data = self._client.get(path).json()

        return from_api_representation_to_job_status_enum(data["jobStatus"])

    def download_error_report(self, download_file_path: Path) -> None:
        """Download the error report of the OCR job associated with this OCR job resource.

        Parameters
        ----------
        download_file_path: Path
            path to download error report

        Returns
        -------
        None
        """
        path = f"{self._path}{self.id}/errorReport/"
        response = self._client.get(path, stream=True)
        for chunk in response.iter_content(chunk_size=1024 * 1024):
            download_file_path.write_bytes(chunk)



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from io import StringIO
from typing import Any, cast, List, Optional, TYPE_CHECKING, Union
from urllib.parse import urlencode

import pandas as pd
import trafaret as t

from datarobot import enums, errors
from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject
from datarobot.utils import from_api, parse_time, raw_prediction_response_to_dataframe

_base_metadata_path = "projects/{project_id}/predictionsMetadata/"
_get_metadata_path = _base_metadata_path + "{prediction_id}/"
# get predictions for particular prediction id.
_get_path = "projects/{project_id}/predictions/{prediction_id}/"

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    from datarobot.models.api_object import ServerDataDictType

    class ShapWarnings(TypedDict):
        mismatch_row_count: int
        max_normalized_mismatch: float

    class PredictionAsDict(TypedDict):  # pylint: disable=missing-class-docstring
        id: str
        model_id: str
        prediction_dataset_id: Optional[str]
        dataset_id: Optional[str]
        includes_prediction_intervals: bool
        prediction_intervals_size: Optional[int]
        forecast_point: Optional[datetime]
        predictions_start_date: Optional[datetime]
        predictions_end_date: Optional[datetime]
        actual_value_column: Optional[str]
        explanation_algorithm: Optional[str]
        max_explanations: Optional[int]
        shap_warnings: Optional[ShapWarnings]

    class PredictionsResponse(TypedDict):
        predictions: PredictionAsDict
        task: Any  # TODO - Annotate this better
        shapBaseValue: List[float]


class Predictions(APIObject):
    """
    Represents predictions metadata and provides access to prediction results.

    Attributes
    ----------
    project_id : str
        id of the project the model belongs to
    model_id : str
        id of the model
    prediction_id : str
        id of generated predictions
    includes_prediction_intervals : Optional[bool]
        (New in v2.16) For :ref:`time series <time-series>` projects only.
        Indicates if prediction intervals will be part of the response. Defaults to False.
    prediction_intervals_size : Optional[int]
        (New in v2.16) For :ref:`time series <time-series>` projects only.
        Indicates the percentile used for prediction intervals calculation. Will be present only
        if `includes_prediction_intervals` is True.
    forecast_point : datetime.datetime, optional
        (New in v2.20) For :ref:`time series <time-series>` projects only. This is the default point
        relative to which predictions will be generated, based on the forecast window of the
        project. See the time series :ref:`prediction documentation <time-series-predict>` for more
        information.
    predictions_start_date : datetime.datetime or None, optional
        (New in v2.20) For :ref:`time series <time-series>` projects only. The start date for bulk
        predictions. Note that this parameter is for generating historical predictions using the
        training data. This parameter should be provided in conjunction with
        ``predictions_end_date``. Can't be provided with the ``forecast_point`` parameter.
    predictions_end_date : datetime.datetime or None, optional
        (New in v2.20) For :ref:`time series <time-series>` projects only. The end date for bulk
        predictions, exclusive. Note that this parameter is for generating historical predictions
        using the training data. This parameter should be provided in conjunction with
        ``predictions_start_date``. Can't be provided with the ``forecast_point`` parameter.
    actual_value_column : string, optional
        (New in version v2.21) For :ref:`time series <time-series>` unsupervised projects only.
        Actual value column which was used to calculate the classification metrics and
        insights on the prediction dataset. Can't be provided with the ``forecast_point``
        parameter.
    explanation_algorithm : datarobot.enums.EXPLANATIONS_ALGORITHM, optional
        (New in version v2.21) If set to 'shap', the response will include prediction
        explanations based on the SHAP explainer (SHapley Additive exPlanations). Defaults to null
        (no prediction explanations).
    max_explanations : Optional[int]
        (New in version v2.21) The maximum number of explanation values that should be returned
        for each row, ordered by absolute value, greatest to least. If null, no limit. In the case
        of 'shap': if the number of features is greater than the limit, the sum of remaining values
        will also be returned as `shapRemainingTotal`. Defaults to null. Cannot be set if
        `explanation_algorithm` is omitted.
    shap_warnings : dict, optional
        (New in version v2.21) Will be present if `explanation_algorithm` was set to
        `datarobot.enums.EXPLANATIONS_ALGORITHM.SHAP` and there were additivity failures during SHAP
        values calculation.


    Examples
    --------

    List all predictions for a project

    .. code-block:: python

        import datarobot as dr

        # Fetch all predictions for a project
        all_predictions = dr.Predictions.list(project_id)

        # Inspect all calculated predictions
        for predictions in all_predictions:
            print(predictions)  # repr includes project_id, model_id, and dataset_id

    Retrieve predictions by id

    .. code-block:: python

        import datarobot as dr

        # Getting predictions by id
        predictions = dr.Predictions.get(project_id, prediction_id)

        # Dump actual predictions
        df = predictions.get_all_as_dataframe()
        print(df)
    """

    def __init__(
        self,
        project_id: str,
        prediction_id: str,
        model_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
        includes_prediction_intervals: Optional[bool] = None,
        prediction_intervals_size: Optional[int] = None,
        forecast_point: Optional[datetime] = None,
        predictions_start_date: Optional[datetime] = None,
        predictions_end_date: Optional[datetime] = None,
        actual_value_column: Optional[str] = None,
        explanation_algorithm: Optional[str] = None,
        max_explanations: Optional[int] = None,
        shap_warnings: Optional[ShapWarnings] = None,
    ) -> None:
        self.project_id = project_id
        self.model_id = model_id
        self.dataset_id = dataset_id
        self.prediction_id = prediction_id
        self.path = _get_path.format(project_id=self.project_id, prediction_id=self.prediction_id)
        self.includes_prediction_intervals = includes_prediction_intervals
        self.prediction_intervals_size = prediction_intervals_size
        self.forecast_point = forecast_point
        self.predictions_start_date = predictions_start_date
        self.predictions_end_date = predictions_end_date
        self.actual_value_column = actual_value_column
        self.explanation_algorithm = explanation_algorithm
        self.max_explanations = max_explanations
        self.shap_warnings = shap_warnings

    _metadata_trafaret = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("url"): String(),
            t.Key("model_id"): String(),
            t.Key("prediction_dataset_id"): String(),
            t.Key("includes_prediction_intervals"): t.Bool(),
            t.Key("prediction_intervals_size", optional=True): Int(),
            t.Key("forecast_point", optional=True): parse_time,
            t.Key("predictions_start_date", optional=True): parse_time,
            t.Key("predictions_end_date", optional=True): parse_time,
            t.Key("actual_value_column", optional=True): String(),
            t.Key("explanation_algorithm", optional=True): String(),
            t.Key("max_explanations", optional=True): Int(),
            t.Key("shap_warnings", optional=True): t.Dict(
                {t.Key("mismatch_row_count"): Int(), t.Key("max_normalized_mismatch"): t.Float()}
            ),
        }
    ).ignore_extra("*")

    @classmethod
    def _build_list_path(  # pylint: disable=missing-function-docstring
        cls, project_id: str, model_id: Optional[str] = None, dataset_id: Optional[str] = None
    ) -> str:
        args = {}
        if model_id:
            args["modelId"] = model_id
        if dataset_id:
            args["predictionDatasetId"] = dataset_id

        path = _base_metadata_path.format(project_id=project_id)
        if args:
            path = f"{path}?{urlencode(args)}"

        return path

    @classmethod
    def _from_server_object(  # pylint: disable=missing-function-docstring
        cls,
        project_id: str,
        item: PredictionAsDict,
    ) -> Predictions:
        pred = cls(
            project_id,
            prediction_id=item["id"],
            model_id=item["model_id"],
            dataset_id=item.get("prediction_dataset_id") or item.get("dataset_id"),
            includes_prediction_intervals=item["includes_prediction_intervals"],
        )
        if pred.includes_prediction_intervals:
            pred.prediction_intervals_size = item["prediction_intervals_size"]
        if item.get("forecast_point"):
            pred.forecast_point = item["forecast_point"]
        if item.get("predictions_start_date"):
            pred.predictions_start_date = item["predictions_start_date"]
        if item.get("predictions_end_date"):
            pred.predictions_end_date = item["predictions_end_date"]
        if item.get("actual_value_column"):
            pred.actual_value_column = item["actual_value_column"]
        if item.get("explanation_algorithm"):
            pred.explanation_algorithm = item["explanation_algorithm"]
        if item.get("max_explanations"):
            pred.max_explanations = item["max_explanations"]
        if item.get("shap_warnings"):
            pred.shap_warnings = item["shap_warnings"]

        return pred

    @classmethod
    def list(
        cls, project_id: str, model_id: Optional[str] = None, dataset_id: Optional[str] = None
    ) -> List[Predictions]:
        """
        Fetch all the computed predictions metadata for a project.

        Parameters
        ----------
        project_id : str
            id of the project
        model_id : Optional[str]
            if specified, only predictions metadata for this model will be retrieved
        dataset_id : Optional[str]
            if specified, only predictions metadata for this dataset will be retrieved

        Returns
        -------
        A list of :py:class:`Predictions <datarobot.models.Predictions>` objects
        """
        path = cls._build_list_path(project_id, model_id=model_id, dataset_id=dataset_id)
        converted = from_api(cls._server_data(path))
        retval = []
        for item in cast("ServerDataDictType", converted)["data"]:
            validated = cls._metadata_trafaret.check(item)
            pred = cls._from_server_object(project_id, validated)
            retval.append(pred)
        return retval

    @classmethod
    def get(cls, project_id: str, prediction_id: str) -> Predictions:
        """
        Retrieve the specific predictions metadata

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        prediction_id : str
            id of the prediction set

        Returns
        -------
        :py:class:`Predictions <datarobot.models.Predictions>` object representing specified
        predictions
        """
        path = _get_metadata_path.format(project_id=project_id, prediction_id=prediction_id)

        converted = from_api(cls._server_data(path))
        validated = cls._metadata_trafaret.check(converted)

        return cls._from_server_object(project_id, validated)

    def get_all_as_dataframe(
        self, class_prefix: str = enums.PREDICTION_PREFIX.DEFAULT, serializer: str = "json"
    ) -> pd.DataFrame:
        """
        Retrieve all prediction rows and return them as a pandas.DataFrame.

        Parameters
        ----------
        class_prefix : Optional[str]
            The prefix to append to labels in the final dataframe. Default is ``class_``
            (e.g., apple -> class_apple)
        serializer : Optional[str]
            Serializer to use for the download. Options: ``json`` (default) or ``csv``.

        Returns
        -------
        dataframe: pandas.DataFrame

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        serializers = {
            "json": self._get_all_as_dataframe_json,
            "csv": self._get_all_as_dataframe_csv,
        }
        if serializer not in serializers:
            raise ValueError(f'Unknown serializer "{serializer}", use "json" or "csv"')

        return serializers[serializer](class_prefix)

    def _get_all_as_dataframe_json(self, class_prefix: str) -> pd.DataFrame:
        data = self._server_data(self.path)
        return raw_prediction_response_to_dataframe(cast("PredictionsResponse", data), class_prefix)

    def _get_all_as_dataframe_csv(
        self,
        class_prefix: str,  # pylint: disable=unused-argument
    ) -> pd.DataFrame:
        resp = self._client.get(self.path, headers={"Accept": "text/csv"}, stream=True)
        if resp.status_code == 200:
            content = resp.content.decode("utf-8")
            return pd.read_csv(StringIO(content), encoding="utf-8")
        else:
            raise errors.ServerError(
                f"Server returned unknown status code: {resp.status_code}",
                resp.status_code,
            )

    def download_to_csv(
        self, filename: Union[str, bool], encoding: str = "utf-8", serializer: str = "json"
    ) -> None:
        """
        Save prediction rows into CSV file.

        Parameters
        ----------
        filename : str or file object
            path or file object to save prediction rows
        encoding : string, optional
            A string representing the encoding to use in the output file, defaults to
            'utf-8'
        serializer : Optional[str]
            Serializer to use for the download. Options: ``json`` (default) or ``csv``.
        """
        df = self.get_all_as_dataframe(serializer=serializer)
        df.to_csv(
            path_or_buf=filename,
            header=True,
            index=False,
            encoding=encoding,
        )

    def __repr__(self) -> str:
        template = (
            "{}(prediction_id={!r}, project_id={!r}, model_id={!r}, dataset_id={!r}, "
            "includes_prediction_intervals={!r}, prediction_intervals_size={!r}, "
            "forecast_point={!r}, predictions_start_date={!r}, "
            "predictions_end_date={!r}, actual_value_column={!r}, "
            "explanation_algorithm={!r}, max_explanations={!r}, shap_warnings={!r})"
        )
        return template.format(
            type(self).__name__,
            self.prediction_id,
            self.project_id,
            self.model_id,
            self.dataset_id,
            self.includes_prediction_intervals,
            self.prediction_intervals_size,
            self.forecast_point,
            self.predictions_start_date,
            self.predictions_end_date,
            self.actual_value_column,
            self.explanation_algorithm,
            self.max_explanations,
            self.shap_warnings,
        )



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot.enums import AggregationType, enum_to_list, GradingResult
from datarobot.models.api_object import APIObject
from datarobot.models.genai.evaluation_dataset_metric_aggregation import aggregation_value_trafaret
from datarobot.models.genai.llm_blueprint import LLMBlueprint
from datarobot.models.genai.llm_test_configuration import (
    insight_grading_criteria_trafaret,
    InsightGradingCriteria,
    llm_test_grading_criteria_trafaret,
    LLMTestConfiguration,
    LLMTestGradingCriteria,
)
from datarobot.utils.pagination import unpaginate


def get_entity_id(entity: Union[LLMTestConfiguration, LLMBlueprint, LLMTestResult, str]) -> str:
    """
    Get the entity ID from the entity parameter.

    Parameters
    ----------
    entity : ApiObject or str
        May be entity ID or the entity.

    Returns
    -------
    id : str
        Entity ID
    """
    if isinstance(entity, str):
        return entity

    return entity.id


insight_evaluation_result_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("llm_test_result_id"): t.String,
        t.Key("evaluation_dataset_configuration_id", optional=True): t.Or(t.String, t.Null),
        t.Key("evaluation_dataset_name"): t.String(allow_blank=True),
        t.Key("metric_name", optional=True): t.Or(t.String, t.Null),
        t.Key("chat_id"): t.String,
        t.Key("chat_name"): t.String(allow_blank=True),
        t.Key("aggregation_value", optional=True): t.Or(
            t.Float, t.List(aggregation_value_trafaret), t.Null
        ),
        t.Key("aggregation_type", optional=True): t.Or(
            t.Enum(*enum_to_list(AggregationType)), t.Null
        ),
        t.Key("grading_result", optional=True): t.Or(t.Enum(*enum_to_list(GradingResult)), t.Null),
        t.Key("execution_status"): t.String,
        t.Key("evaluation_name"): t.String(allow_blank=True),
        t.Key("insight_grading_criteria"): insight_grading_criteria_trafaret,
        t.Key("last_update_date"): t.String,
    }
).ignore_extra("*")


llm_test_result_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("llm_test_configuration_id"): t.String,
        t.Key("llm_test_configuration_name"): t.String(allow_blank=True),
        t.Key("use_case_id"): t.String,
        t.Key("llm_blueprint_id"): t.String,
        t.Key("llm_blueprint_snapshot"): t.Dict().allow_extra("*"),
        t.Key("llm_test_grading_criteria"): llm_test_grading_criteria_trafaret,
        t.Key("grading_result", optional=True): t.Or(t.Enum(*enum_to_list(GradingResult)), t.Null),
        t.Key("pass_percentage", optional=True): t.Or(t.Float, t.Null),
        t.Key("execution_status"): t.String,
        t.Key("insight_evaluation_results"): t.List(insight_evaluation_result_trafaret),
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
        t.Key("creation_user_name"): t.String(allow_blank=True),
    }
).ignore_extra("*")


class InsightEvaluationResult(APIObject):
    """
    Metadata for a DataRobot GenAI insight evaluation result.

    Attributes
    ----------
    id : str
        The ID of the insight evaluation result.
    llm_test_result_id : str
        The ID of the LLM test result associated with this insight evaluation result.
    evaluation_dataset_configuration_id : str
        The ID of the evaluation dataset configuration.
    evaluation_dataset_name : str
        The name of the evaluation dataset.
    metric_name : str
        The name of the metric.
    chat_id : str
        The ID of the chat containing the prompts and responses.
    chat_name : str
        The name of the chat containing the prompts and responses.
    aggregation_type : AggregationType
        The type of aggregation used for the metric results.
    grading_result : GradingResult
        The overall grade for the LLM test.
    execution_status : str
        The execution status of the LLM test.
    evaluation_name : str
        The name of the evaluation.
    insight_grading_criteria: InsightGradingCriteria
        The criteria to grade the results.
    last_update_date : str
        The date the result was most recently updated.
    aggregation_value : float | List[Dict[str, float]] | None
        The aggregated metric result.
    """

    _converter = insight_evaluation_result_trafaret

    def __init__(
        self,
        id: str,
        llm_test_result_id: str,
        chat_id: str,
        chat_name: str,
        execution_status: str,
        evaluation_name: str,
        insight_grading_criteria: Dict[str, Any],
        last_update_date: str,
        evaluation_dataset_name: str,
        aggregation_value: Union[float, List[Dict[str, float]], None] = None,
        evaluation_dataset_configuration_id: Optional[str] = None,
        metric_name: Optional[str] = None,
        aggregation_type: Optional[AggregationType] = None,
        grading_result: Optional[GradingResult] = None,
    ):
        self.id = id
        self.llm_test_result_id = llm_test_result_id
        self.evaluation_dataset_configuration_id = evaluation_dataset_configuration_id
        self.metric_name = metric_name
        self.chat_id = chat_id
        self.chat_name = chat_name
        self.aggregation_type = aggregation_type
        self.grading_result = grading_result
        self.execution_status = execution_status
        self.evaluation_name = evaluation_name
        self.evaluation_dataset_name = evaluation_dataset_name
        self.insight_grading_criteria = InsightGradingCriteria.from_server_data(
            insight_grading_criteria
        )
        self.last_update_date = last_update_date
        self.aggregation_value = aggregation_value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id})"


class LLMTestResult(APIObject):
    """
    Metadata for a DataRobot GenAI LLM test result.

    Attributes
    ----------
    id : str
        The LLM test result ID.
    llm_test_configuration_id : str
        The LLM test configuration ID associated with this LLM test result.
    llm_test_configuration_name : str
        The LLM test configuration name associated with this LLM test result.
    use_case_id : str
        The ID of the Use Case associated with this LLM test result.
    llm_blueprint_id : str
        The ID of the LLM blueprint for this LLM test result.
    llm_test_grading_criteria : LLMTestGradingCriteria
        The criteria used to grade the result of the LLM test configuration.
    grading_result : GradingResult
        The overall grading result for the LLM test.
    pass_percentage : float
        The percentage of insight evaluation results that passed the grading criteria.
    execution_status : str
        The execution status of the job that evaluated the LLM test result.
    insight_evaluation_result : list[InsightEvaluationResult]
        The results for the individual insights that make up the LLM test result.
    creation_date : str
        The date of the LLM test result.
    creation_user_id : str
        The ID of the user who executed the LLM test.
    creation_user_name: str
        The name of the user who executed the LLM test.
    """

    _path = "api/v2/genai/llmTestResults"

    _converter = llm_test_result_trafaret

    def __init__(
        self,
        id: str,
        llm_test_configuration_id: str,
        llm_test_configuration_name: str,
        use_case_id: str,
        llm_blueprint_id: str,
        llm_blueprint_snapshot: Dict[str, Any],
        llm_test_grading_criteria: Dict[str, Any],
        execution_status: str,
        insight_evaluation_results: List[Dict[str, Any]],
        creation_date: str,
        creation_user_id: str,
        creation_user_name: str,
        pass_percentage: Optional[float] = None,
        grading_result: Optional[GradingResult] = None,
    ):
        self.id = id
        self.llm_test_configuration_id = llm_test_configuration_id
        self.llm_test_configuration_name = llm_test_configuration_name
        self.use_case_id = use_case_id
        self.llm_blueprint_id = llm_blueprint_id
        self.llm_test_grading_criteria = LLMTestGradingCriteria.from_server_data(
            llm_test_grading_criteria
        )
        self.pass_percentage = pass_percentage
        self.execution_status = execution_status
        self.llm_blueprint_snapshot = llm_blueprint_snapshot
        self.insight_evaluation_results = [
            InsightEvaluationResult.from_server_data(result)
            for result in insight_evaluation_results
        ]
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.creation_user_name = creation_user_name
        self.grading_result = grading_result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id})"

    @classmethod
    def create(
        cls,
        llm_test_configuration: Union[LLMTestConfiguration, str],
        llm_blueprint: Union[LLMBlueprint, str],
    ) -> LLMTestResult:
        """
        Create a new LLMTestResult. This executes the LLM test configuration using the
        specified LLM blueprint. To check the status of the LLM test, use the
        `LLMTestResult.get` method with the returned ID.

        Parameters
        ----------
        llm_test_configuration : LLMTestConfiguration or str
            The LLM test configuration to execute, either `LLMTestConfiguration` or
            the LLM test configuration ID.
        llm_blueprint : LLMBlueprint or str
            The LLM blueprint to test, either `LLMBlueprint` or
            the LLM blueprint ID.

        Returns
        -------
        llm_test_result : LLMTestResult
            The created LLM test result.
        """
        payload = {
            "llm_test_configuration_id": get_entity_id(llm_test_configuration),
            "llm_blueprint_id": get_entity_id(llm_blueprint),
        }

        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, json=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def get(cls, llm_test_result: Union[LLMTestResult, str]) -> LLMTestResult:
        """
        Retrieve a single LLM test result.

        Parameters
        ----------
        llm_test_result : LLMTestResult or str
            The LLM test result to retrieve, specified by either LLM test result or test ID.

        Returns
        -------
        llm_test_result : LLMTestResult
            The requested LLM test result.
        """
        url = f"{cls._client.domain}/{cls._path}/{get_entity_id(llm_test_result)}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        llm_test_configuration: Optional[Union[LLMTestConfiguration, str]] = None,
        llm_blueprint: Optional[Union[LLMBlueprint, str]] = None,
    ) -> List[LLMTestResult]:
        """
        List all LLM test results available to the user. If the LLM test configuration or LLM
        blueprint is specified, results are restricted to only those LLM test results associated
        with the LLM test configuration or LLM blueprint.

        Parameters
        ----------
        llm_test_configuration : Optional[Union[LLMTestConfiguration, str]]
            The returned LLM test results are filtered to those associated with a specific
            LLM test configuration, if specified.
        llm_blueprint : Optional[Union[LLMBlueprint, str]]
            The returned LLM test results, filtered by those associated with a specific
            LLM blueprint, if specified.

        Returns
        -------
        llm_test_results : List[LLMTestResult]
            Returns a list of LLM test results.
        """

        params = {}
        if llm_test_configuration:
            params["llm_test_configuration_id"] = get_entity_id(llm_test_configuration)
        if llm_blueprint:
            params["llm_blueprint_id"] = get_entity_id(llm_blueprint)
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def delete(self) -> None:
        """
        Delete a single LLM test result.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)
